import gzip
import io
import os

from collections import namedtuple, deque

from six import string_types as basestring

OpenerRecord = namedtuple("OpenerRecord", ('opener', 'extension', 'magic_bytes'))

DEFAULT_OPENER = OpenerRecord(open, None, None)


class OpenerRegistry(object):
    def __init__(self, openers=None, default=DEFAULT_OPENER):
        if openers is None:
            openers = []
        self.openers = deque(openers)
        self.default = default

    def add(self, opener, extension=None, magic_bytes=None):
        record = OpenerRecord(opener, extension, magic_bytes)
        self.openers.appendleft(record)

    def __iter__(self):
        return iter(self.openers)

    def __len__(self):
        return len(self.openers)

    def __getitem__(self, i):
        return self.openers[i]

    def by_extension(self, extension):
        for rec in self:
            if rec.extension is None:
                continue
            if rec.extension == extension:
                return rec
        return None

    def by_magic_bytes(self, bytestring):
        for rec in self:
            if rec.magic_bytes is None:
                continue
            if bytestring.startswith(rec.magic_bytes):
                return rec
        return None

    def get(self, path):
        opener = None
        if isinstance(path, basestring):
            ext = os.path.splitext(path)[1]
            opener = self.by_extension(ext)
            if opener is None:
                with open(path, 'rb') as fh:
                    bytestring = fh.read(100)
                    opener = self.by_magic_bytes(bytestring)
        elif hasattr(path, 'read') and hasattr(path, 'tell'):
            current = path.tell()
            path.seek(0)
            bytestring = path.read(100)
            path.seek(current)
            opener = self.by_magic_bytes(bytestring)
        if opener is None:
            return self.default.opener
        else:
            return opener.opener


openers = OpenerRegistry([
    OpenerRecord(gzip.GzipFile, 'gz', b'\037\213'),
])


try:
    import bz2
    openers.add(bz2.BZ2File, 'bz2', None)
except ImportError:
    pass

register = openers.add
get = openers.get



import itertools
import gzip
import os
import tempfile

from psims.mzml import MzMLWriter, binary_encoding
from pyteomics import mzml
import numpy as np
from lxml import etree

import pytest

from psims import compression as compression_registry
from psims.test.utils import output_path, compressor


mz_array = [
    255.22935009, 283.26141863, 284.26105318, 301.23572871,
    304.908247, 329.26327093, 755.25267878, 910.6317435,
    960.96747136, 971.31396162, 972.649568, 991.66036894,
    1017.87649113, 1060.29899182, 1112.67519902, 1113.11545762,
    1113.86200673, 1114.42377982, 1152.34596544, 1177.73119994,
    1188.36935517, 1214.70161813, 1265.9795606, 1266.16111855,
    1293.14606767, 1294.68263447, 1367.01605133, 1565.95282753,
    1700.23290184, 969.65408783, 1110.37794027, 1170.89893785,
    1175.34669421, 1183.40737076, 861.61958381, 1292.94207114,
    1295.4429046, 876.96085225, 1335.39755355, 1357.92354342,
    1365.96972386, 925.64504217, 958.65480011, 1438.49014079,
    1452.48402739, 967.98859816, 986.63557879, 1480.46326372,
    1001.97032019, 1007.67089513, 1525.51932956, 1016.67747057,
    1080.3583722, 1090.03199733, 1133.01762219, 1186.72735997,
    960.79647487, 1274.44354121, 1918.2693496, 961.85503396
]

intensity_array = [
    1.90348869e+03, 1.92160377e+03, 3.26032338e+02,
    1.05527732e+03, 9.50991606e+02, 1.52403574e+03,
    8.63154395e+02, 4.11169655e+02, 2.33462730e+03,
    2.62603673e+02, 2.73669694e+02, 8.62436899e+02,
    4.22323174e+02, 2.54371429e+02, 1.02364420e+03,
    5.44244205e+02, 4.93101348e+02, 2.64984906e+02,
    9.36500725e+02, 4.79373626e+02, 9.26742857e+02,
    4.52209221e+02, 3.02178809e+03, 4.94385979e+02,
    1.67240655e+03, 9.41320838e+02, 7.25744090e+02,
    1.27260012e+03, 8.24236545e+02, 4.93518583e+02,
    7.33281806e+04, 9.60817582e+02, 2.64893187e+03,
    7.95614286e+03, 4.94586514e+03, 2.35346153e+04,
    1.50301526e+03, 5.36435167e+03, 3.78042332e+02,
    1.09926345e+03, 3.10133857e+03, 7.41566590e+02,
    2.77340229e+05, 1.00796887e+05, 4.69519356e+03,
    1.55343822e+04, 5.45621612e+03, 5.53939031e+03,
    9.49732490e+03, 8.05000735e+03, 2.65457068e+03,
    1.36766228e+04, 2.69348480e+03, 6.71802368e+03,
    4.46828571e+02, 1.39065143e+04, 4.29267365e+03,
    2.73782365e+03, 1.35373492e+03, 1.17601397e+03
]

charge_array = [
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2,
    -3, -2, -2, -3, -2, -2, -2, -3, -3, -2, -2, -3, -3, -2, -3, -3, -2,
    -3, -3, -3, -3, -3, -5, -4, -3, -6
]


sample = {
    "name": "shotgun_explodeomics_precipitate_123",
    "params": [
        "certified organic",
        {"acquired in": "1812"}
    ]
}

encodings = {
    "m/z array": np.float64,
    "intensity array": np.float64,
    "charge array": np.float64
}


def test_array_codec():
    encode = binary_encoding.encode_array
    decode = binary_encoding.decode_array

    compression_types = [binary_encoding.COMPRESSION_ZLIB, binary_encoding.COMPRESSION_NONE]

    dtypes = [np.float64, np.float32]

    for dtype, compression in itertools.product(dtypes, compression_types):
        original = np.array(intensity_array, dtype=dtype)
        encoded = encode(original, compression, dtype)
        decoded = decode(encoded, compression, dtype)
        np.allclose(original, decoded)


def test_write(output_path, compressor):
    f = MzMLWriter(compressor(output_path, 'wb'), close=True)
    f.register("Software", 'psims')
    with f:
        f.controlled_vocabularies()
        f.file_description(["spam", "MS1 spectrum", "MSn spectrum"], [
            dict(id="SPAM1", name="Spam.raw", location="file:///", params=[dict(name="Thermo RAW format")])])
        f.reference_param_group_list([
            {'id': 'common_params', 'params': [{"proven": "inductively"}]}
        ])
        f.sample_list([sample])
        f.software_list([
            f.Software(version="0.0.0", id='psims', params=['custom unreleased software tool', 'psims'])
        ])
        f.instrument_configuration_list([
            f.InstrumentConfiguration(id=1, component_list=f.ComponentList([
                f.Source(params=['electrospray ionization'], order=1),
                f.Analyzer(params=['quadrupole'], order=2),
                f.Detector(params=['inductive detector'], order=3)
            ])),
            f.InstrumentConfiguration(id=2, component_list=f.ComponentList([
                f.Source(params=['electrospray ionization'], order=1),
                f.Analyzer(params=['radial ejection linear ion trap'], order=2),
                f.Detector(params=['inductive detector'], order=3)
            ]))
        ])
        f.data_processing_list([
            f.DataProcessing(processing_methods=[
                dict(order=0, software_reference='psims', params=['Conversion to mzML'])
            ])
        ])
        with f.run(id='test'):
            with f.spectrum_list(count=2):
                f.write_spectrum(mz_array, intensity_array, charge_array, id='scanId=1', params=[
                    {"name": "ms level", "value": 1}, {"ref": 'common_params'}],
                    polarity='negative scan', encoding=encodings,
                    compression='zlib')
                f.write_spectrum(mz_array, intensity_array, charge_array, id='scanId=2', params=[
                    {"name": "ms level", "value": 2}, {"ref": 'common_params'}],
                    polarity='negative scan', precursor_information={
                        "mz": 1230, "intensity": None, "charge": None, "params": [
                            "no peak detected"
                        ],
                        "scan_id": "scanId=1", "activation": ["collision-induced dissociation",
                                                              {"collision energy": 56.}]
                }, instrument_configuration_id=2, encoding=encodings, compression='zlib')
    try:
        f.format()
    except OSError:
        pass

    output_path = f.outfile.name
    opener = compression_registry.get(output_path)
    assert opener == compressor
    reader = mzml.read(opener(output_path, 'rb'))

    def reset():
        reader.reset()
        reader.seek(0)

    reset()
    sample_data = next(reader.iterfind("sample"))
    assert sample_data['name'] == "shotgun_explodeomics_precipitate_123"
    reset()
    spec = next(reader)
    assert (all(np.abs(spec['m/z array'] - mz_array) < 1e-4))
    assert "negative scan" in spec
    # appears to be broken in reader
    # assert "referenceableParamGroupRef" in spec and spec['referenceableParamGroupRef'][0]['ref'] == 'common_params'
    assert spec['ms level'] == 1
    spec = next(reader)
    assert "negative scan" in spec
    assert spec['ms level'] == 2
    scan_list_struct = spec['scanList']
    reference = None
    for scan in scan_list_struct.get("scan", []):
        reference = scan.get("instrumentConfigurationRef")
        if reference is None:
            continue
    assert reference == "INSTRUMENTCONFIGURATION_2"

    reset()
    inst_config = next(reader.iterfind("instrumentConfiguration"))
    assert inst_config['id'] == 'INSTRUMENTCONFIGURATION_1'
    assert ("quadrupole") in inst_config['componentList']['analyzer'][0]

    reset()
    file_description = next(reader.iterfind("fileDescription"))
    content = file_description['fileContent']
    assert "MS1 spectrum" in content
    assert "spam" in content

    reset()
    run = next(reader.iterfind("run"))
    assert run.get("defaultInstrumentConfigurationRef") == "INSTRUMENTCONFIGURATION_1"

    reset()
    index_list = list(reader.iterfind("index"))
    assert len(index_list) == 1
    assert index_list[0]['name'] == 'spectrum'
    reset()
    spectrum_offsets = list(reader.iterfind("offset"))
    offset = int(spectrum_offsets[0]['offset'])
    reader.seek(offset)
    bytestring = reader.read(100)
    try:
        content = bytestring.decode("utf8")
    except AttributeError:
        content = bytestring
    assert 'index="0"' in content
    is_valid, schema = f.validate()
    assert is_valid, schema.error_log
    return f



from lxml import etree
from six import add_metaclass, string_types as basestring
from gzip import GzipFile

try:
    from collections import Iterable, Mapping
except ImportError:
    from collections.abc import Iterable, Mapping

from psims import compression


def ensure_iterable(obj):
    if obj is None:
        return tuple()
    if not isinstance(obj, Iterable) or isinstance(obj, basestring) or isinstance(obj, Mapping):
        return [obj]
    return obj


def pretty_xml(path, outpath=None, encoding=b'utf-8'):
    tree = etree.parse(path)
    if outpath is None:
        opener = compression.get(path)
        outpath = opener(path, 'wb')
    if hasattr(outpath, 'write'):
        outstream = outpath
    else:
        opener = compression.get(path)
        outstream = opener(outpath, 'wb')
    with outstream:
        outstream.write(b'<?xml version="1.0" encoding="' + encoding + b'"?>\n')
        outstream.write(
            etree.tostring(tree, pretty_print=True))


def simple_repr(self):  # pragma: no cover
    template = "{self.__class__.__name__}({d})"

    def formatvalue(v):
        if isinstance(v, float):
            return "%0.4f" % v
        else:
            return str(v)

    if not hasattr(self, "__slots__"):
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                self.__dict__.items(), key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]
    else:
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                [(name, getattr(self, name)) for name in self.__slots__], key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v)) and not (v is None)]

    return template.format(self=self, d=', '.join(d))



import os
import pkg_resources
try:
    from urllib2 import urlopen, URLError
except ImportError:
    from urllib.request import urlopen, URLError
from .obo import OBOParser
from . import unimod


def _use_vendored_psims_obo():
    return pkg_resources.resource_stream(__name__, "vendor/psi-ms.obo")


def _use_vendored_unit_obo():
    return pkg_resources.resource_stream(__name__, "vendor/unit.obo")


def _use_vendored_pato_obo():
    return pkg_resources.resource_stream(__name__, "vendor/pato.obo")


def _use_vendored_unimod_xml():
    return pkg_resources.resource_stream(__name__, "vendor/unimod_tables.xml")


fallback = {
    ("http://psidev.cvs.sourceforge.net/*checkout*/"
     "psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo"): _use_vendored_psims_obo,
    ("http://psidev.cvs.sourceforge.net/viewvc/*checkout*/"
     "psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo"): _use_vendored_psims_obo,
    ("https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo"): _use_vendored_psims_obo,
    ("http://obo.cvs.sourceforge.net/*checkout*/"
     "obo/obo/ontology/phenotype/unit.obo"): _use_vendored_unit_obo,
    ("http://ontologies.berkeleybop.org/uo.obo"): _use_vendored_unit_obo,
    ("http://ontologies.berkeleybop.org/pato.obo"): _use_vendored_pato_obo,
}


class ControlledVocabulary(object):
    """A Controlled Vocabulary is a collection
    of terms or entities with controlled meanings
    and semantics.

    This object makes entities resolvable by name,
    accession number, or synonym.

    Attributes
    ----------
    id : str
        Unique identifier for this collection
    """
    @classmethod
    def from_obo(cls, handle):
        parser = OBOParser(handle)
        inst = cls(parser.terms, metadata=parser.header, version=parser.version, name=parser.name)
        if len(parser.terms) == 0:
            raise ValueError("Empty Vocabulary")
        return inst

    def __init__(self, terms, id=None, metadata=None, version=None, name=None):
        if metadata is None:
            metadata = dict()
        self.version = version
        self.name = name
        self.terms = terms
        for term in terms.values():
            term.vocabulary = self
        self._names = {
            v['name']: v for v in terms.values()
        }
        self._normalized = {
            v['name'].lower(): v['name']
            for v in terms.values()
        }
        self._synonyms = {}
        for term in terms.values():
            if term.get('synonym'):
                for synonym in term.get('synonym'):
                    self._synonyms[synonym.lower()] = term
        self.id = id
        self.metadata = metadata

    def __getitem__(self, key):
        try:
            return self.terms[key]
        except KeyError as e:
            try:
                return self._names[key]
            except KeyError:
                try:
                    return self._names[self.normalize_name(key)]
                except KeyError as e2:
                    try:
                        return self._synonyms[key.lower()]
                    except KeyError:
                        raise KeyError("%s and %s were not found." % (e, e2))

    def __repr__(self):
        template = ("{self.__class__.__name__}(terms={size}, id={self.id}, "
                    "name={self.name}, version={self.version})")
        return template.format(self=self, size=len(self.terms))

    def __iter__(self):
        return iter(self.terms)

    def keys(self):
        return self.terms.keys()

    def names(self):
        return self._names.keys()

    def items(self):
        return self.terms.items()

    def normalize_name(self, name):
        return self._normalized[name.lower()]


class OBOCache(object):
    """A cache for retrieved ontology sources

    Attributes
    ----------
    cache_exists : bool
        Whether the cache directory exists
    cache_path : str
        The path to the cache directory
    enabled : bool
        Whether the cache will be used or not
    resolvers : dict
        A mapping from ontology URL to a function
        which will be called instead of opening the
        URL to retrieve the :class:`ControlledVocabulary`
        object.
    """

    def __init__(self, cache_path='.obo_cache', enabled=True, resolvers=None):
        self._cache_path = None
        self.cache_path = cache_path
        self.enabled = enabled
        self.resolvers = resolvers or {}

    @property
    def cache_path(self):
        return self._cache_path

    @cache_path.setter
    def cache_path(self, value):
        self._cache_path = value
        self.cache_exists = os.path.exists(self.cache_path)

    def path_for(self, name, setext=True):
        if not self.cache_exists:
            os.makedirs(self.cache_path)
            self.cache_exists = True
        name = os.path.basename(name)
        if not name.endswith(".obo") and setext:
            name += '.obo'
        return os.path.join(self.cache_path, name)

    def _open_url(self, uri):
        try:
            f = urlopen(uri)
            code = None
            # The keepalive library monkey patches urllib2's urlopen and returns
            # an object with a different API. First handle the normal case, then
            # the patched case.
            if hasattr(f, 'getcode'):
                code = f.getcode()
            elif hasattr(f, "code"):
                code = f.code
            else:
                raise ValueError("Can't understand how to get HTTP response code from %r" % f)
            if code != 200:
                raise ValueError("%s did not resolve" % uri)
        except Exception:
            if uri in fallback:
                f = fallback[uri]()
            else:
                raise ValueError(uri)
        return f

    def fallback(self, uri):
        if uri in fallback:
            f = fallback[uri]()
        else:
            f = None
        return f

    def resolve(self, uri):
        if uri in self.resolvers:
            return self.resolvers[uri](self)
        try:
            if self.enabled:
                name = self.path_for(uri)
                if os.path.exists(name) and os.path.getsize(name) > 0:
                    return open(name, 'rb')
                else:
                    f = self._open_url(uri)
                    with open(name, 'wb') as cache_f:
                        n_chars = 0
                        for i, line in enumerate(f.readlines()):
                            n_chars += len(line)
                            cache_f.write(line)
                        if n_chars < 5:
                            raise ValueError("No bytes written")
                    if os.path.getsize(name) > 0:
                        return open(name, 'rb')
                    else:
                        raise ValueError("Failed to download .obo")
            else:
                f = self._open_url(uri)
                return f
        except ValueError:
            import traceback
            traceback.print_exc()
            raise

    def set_resolver(self, uri, provider):
        self.resolvers[uri] = provider

    def __repr__(self):
        return "OBOCache(cache_path=%r, enabled=%r, resolvers=%s)" % (
            self.cache_path, self.enabled, self.resolvers)


def _make_relative_sqlite_sqlalchemy_uri(path):
    return "sqlite:///%s" % path


def resolve_unimod(cache):
    if cache.enabled:
        path = _make_relative_sqlite_sqlalchemy_uri(
            cache.path_for("unimod.db", False))
        try:
            return unimod.Unimod(path)
        except IOError:
            return unimod.Unimod(path, _use_vendored_unimod_xml())
    else:
        try:
            return unimod.Unimod()
        except IOError:
            return unimod.Unimod(None, _use_vendored_unimod_xml())


obo_cache = OBOCache(enabled=False)
obo_cache.set_resolver("http://www.unimod.org/obo/unimod.obo", resolve_unimod)


def configure_obo_store(path):
    if path is None:
        obo_cache.enabled = False
    else:
        obo_cache.cache_path = path
        obo_cache.enabled = True


def register_resolver(name, fn):
    obo_cache.set_resolver(name, fn)


def load_psims():
    cv = obo_cache.resolve(("https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo"))
    return ControlledVocabulary.from_obo(cv)


def load_uo():
    cv = obo_cache.resolve("http://ontologies.berkeleybop.org/uo.obo")
    return ControlledVocabulary.from_obo(cv)


def load_pato():
    cv = obo_cache.resolve("http://ontologies.berkeleybop.org/pato.obo")
    return ControlledVocabulary.from_obo(cv)



import warnings
import operator
import re

from collections import Mapping

from datetime import datetime
from numbers import Number as NumberBase
from itertools import chain

from ..xml import (
    _element, element, TagBase,
    CVParam, UserParam, CV,
    ProvidedCV, sanitize_id)
from ..document import (
    ComponentBase as _ComponentBase, NullMap, ComponentDispatcherBase,
    ParameterContainer)

from .utils import ensure_iterable

from lxml import etree


AUTO = object()


class MzIdentML(TagBase):
    v1_1_0_type_attrs = {
        "xmlns": "http://psidev.info/psi/pi/mzIdentML/1.1",
        "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "xsi:schemaLocation": "http://psidev.info/psi/pi/mzIdentML/1.1 ../../schema/mzIdentML1.1.0.xsd"
    }

    v1_2_0_type_attrs = {
        "xmlns": "http://psidev.info/psi/pi/mzIdentML/1.2",
        "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "xsi:schemaLocation": "http://psidev.info/psi/pi/mzIdentML/1.2 ../../schema/mzIdentML1.2.0.xsd"
    }

    def __init__(self, **attrs):
        attrs.setdefault('creationDate', datetime.utcnow().isoformat())
        attrs.setdefault("id", 0)
        attrs.setdefault("version", '1.1.0')
        version = attrs['version']
        if version == '1.1.0':
            self.type_attrs = self.v1_1_0_type_attrs
        elif version == '1.2.0':
            self.type_attrs = self.v1_2_0_type_attrs

        super(MzIdentML, self).__init__("MzIdentML", **attrs)


_COMPONENT_NAMESPACE = 'mzid'


class ComponentDispatcher(ComponentDispatcherBase):
    def __init__(self, *args, **kwargs):
        super(ComponentDispatcher, self).__init__(*args, component_namespace=_COMPONENT_NAMESPACE, **kwargs)


class ComponentBase(_ComponentBase):
    component_namespace = _COMPONENT_NAMESPACE


def _unimod_converter(modification):
    return {
        "name": modification.code_name,
        "id": "UNIMOD:%s" % modification.id,
        "_object": modification
    }


default_cv_list = [
    CV(
        id="PSI-MS",
        uri=("http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev"
             "/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo"),
        full_name="PSI-MS"),
    CV(
        id="UO",
        uri="http://ontologies.berkeleybop.org/uo.obo",
        full_name="UNIT-ONTOLOGY"),
    ProvidedCV(id="UNIMOD", uri="http://www.unimod.org/obo/unimod.obo", full_name="UNIMOD", converter=_unimod_converter)
]


common_units = {
    "parts per million": "UO:0000169",
    "dalton": "UO:0000221"
}


_xmlns = "http://psidev.info/psi/pi/mzIdentML/1.2"


class GenericCollection(ComponentBase):
    def __init__(self, tag_name, members, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, xmlns=_xmlns)

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            for member in self.members:
                member.write(xml_file)


class IDGenericCollection(GenericCollection):
    def __init__(self, tag_name, members, id, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, xmlns=_xmlns, id=id)
        context[tag_name][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            for member in self.members:
                member.write(xml_file)

# --------------------------------------------------
# Input File Information


class SourceFile(ComponentBase):
    def __init__(self, location, file_format, id=None, external_format=None,
                 params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.external_format = external_format
        self.file_format = file_format
        self.element = _element("SourceFile", location=location, id=id)
        self.context = context
        context["SourceFile"][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            with element(xml_file, "FileFormat"):
                self.context.param(self.file_format)(xml_file)
            if self.external_format is not None:
                with element(xml_file, "ExternalFormatDocumentation"):
                    xml_file.write(str(self.external_format))
            self.write_params(xml_file)


class SearchDatabase(ComponentBase):
    def __init__(self, name, file_format, location=None, id=None, external_format=None,
                 params=None, context=NullMap, **kwargs):
        self.external_format = external_format
        self.params = self.prepare_params(params, **kwargs)
        self.location = location
        self.file_format = file_format
        self.element = _element("SearchDatabase", location=location, name=name, id=id)
        context["SearchDatabase"][id] = self.element.id
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            with element(xml_file, "FileFormat"):
                self.write_params(xml_file, self.prepare_params(self.file_format))
            with element(xml_file, "DatabaseName"):
                UserParam(name=self.name).write(xml_file)
            if self.external_format is not None:
                with element(xml_file, "ExternalFormatDocumentation"):
                    xml_file.write(str(self.external_format))
            self.write_params(xml_file)


class SpectraData(ComponentBase):
    def __init__(self, location, file_format, spectrum_id_format, id=None, params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.file_format = file_format
        self.spectrum_id_format = spectrum_id_format
        self.element = _element("SpectraData", id=id, location=location)
        context['SpectraData'][id] = self.element.id
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            with element(xml_file, "FileFormat"):
                self.write_params(xml_file, self.prepare_params(self.file_format))
            with element(xml_file, "SpectrumIDFormat"):
                self.context.param(self.spectrum_id_format)(xml_file)
            self.write_params(xml_file)


class Inputs(GenericCollection):
    def __init__(self, source_files=tuple(), search_databases=tuple(), spectra_data=tuple(), context=NullMap):
        items = list()
        items.extend(source_files)
        items.extend(search_databases)
        items.extend(spectra_data)
        super(Inputs, self).__init__("Inputs", items, context=context)

# --------------------------------------------------
# Search Database Content


class DBSequence(ComponentBase):
    def __init__(self, accession, sequence=None, id=None, search_database_id=1, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.params = params
        self.sequence = sequence
        self.search_database_ref = context['SearchDatabase'][search_database_id]
        self.element = _element(
            "DBSequence", accession=accession, id=id,
            searchDatabase_ref=self.search_database_ref)
        if sequence is not None:
            self.element.attrs["length"] = len(sequence)

        context["DBSequence"][id] = self.element.id
        self.context = context

    def write(self, xml_file):
        protein = self.sequence
        with self.element.element(xml_file, with_id=True):
            if self.sequence is not None:
                with element(xml_file, "Seq"):
                    xml_file.write(protein)
            self.write_params(xml_file)


class Peptide(ComponentBase):
    def __init__(self, peptide_sequence, id, modifications=None, params=None, context=NullMap, **kwargs):
        if modifications is None:
            modifications = []
        params = self.prepare_params(params, **kwargs)
        self.context = context
        self.peptide_sequence = peptide_sequence
        self.modifications = [Modification(context=context, **mod)
                              if not isinstance(mod, Modification) else mod for mod in modifications]
        self.params = params
        self.element = _element("Peptide", id=id)
        context["Peptide"][id] = self.element.id
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            with element(xml_file, "PeptideSequence"):
                xml_file.write(self.peptide_sequence)
            for mod in self.modifications:
                mod.write(xml_file)
            self.write_params(xml_file)


class Modification(ComponentBase):
    def __init__(self, monoisotopic_mass_delta=None, location=None, name=None,
                 id=None, known=True, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        if id is None:
            try:
                mod = context.term(name)
                self.name = mod["name"]
                self.accession = mod['id']
            except KeyError:
                known = False
                self.name = name
                self.accession = "MS:1001460"
        elif name is None:
            try:
                mod = context.term(id)
                self.name = mod["name"]
                self.accession = mod['id']
            except KeyError:
                known = False
                self.name = id
                self.accession = "MS:1001460"
        else:
            warnings.warn("Unknown modification saved: %s" % monoisotopic_mass_delta)
            self.name = name
            self.accession = id

        self.element = _element(
            "Modification", monoisotopicMassDelta=monoisotopic_mass_delta,
            location=location)
        self.context = context
        self.params = params
        self.known = known

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            if self.accession is not None:
                if self.known:
                    self.context.param(
                        name=self.name, accession=self.accession,
                        ref=self.accession.split(":")[0])(xml_file)
                else:
                    psi_ms = self.context.get_vocabulary("PSI-MS")
                    self.context.param(
                        name="unknown modification",
                        accession=self.accession,
                        value=self.name,
                        ref=psi_ms.id if psi_ms is not None else 'PSI-MS')(xml_file)
            self.write_params(xml_file)


class SubstitutionModification(ComponentBase):
    def __init__(self, original_residue, replacement_residue,
                 monoisotopic_mass_delta=None, location=None,
                 context=NullMap):
        self.original_residue = original_residue
        self.replacement_residue = replacement_residue
        self.monoisotopic_mass_delta = monoisotopic_mass_delta
        self.location = location
        self.element = _element(
            "SubstitutionModification", originalResidue=self.original_residue,
            replacementResidue=self.replacement_residue)
        if self.monoisotopic_mass_delta is not None:
            self.element.attrs['monoisotopicMassDelta'] = self.monoisotopic_mass_delta
        if self.location is not None:
            self.element.attrs['location'] = self.location
        self.context = context

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            pass


class PeptideEvidence(ComponentBase):
    def __init__(self, peptide_id, db_sequence_id, id, start_position, end_position,
                 is_decoy=False, pre=None, post=None, params=None, frame=None,
                 translation_table_id=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.params = params
        self.context = context
        self.peptide_id = peptide_id
        self.db_sequence_id = db_sequence_id
        self.translation_table_id = translation_table_id
        self.frame = frame
        self.element = _element(
            "PeptideEvidence", isDecoy=is_decoy, start=start_position,
            end=end_position, peptide_ref=context["Peptide"][peptide_id],
            dBSequence_ref=context['DBSequence'][db_sequence_id],
            pre=pre, post=post, id=id, frame=frame)
        if self.translation_table_id is not None:
            self.element.attrs['translationTable_ref'] = self.context['TranslationTable'][self.translation_table_id]
        self.context["PeptideEvidence"][id] = self.element.id

    def write(self, xml_file):
        if self.params:
            with self.element(xml_file, with_id=True):
                self.write_params(xml_file)
        else:
            xml_file.write(self.element(with_id=True))


# --------------------------------------------------
# Identification Information


class SpectrumIdentificationResult(ComponentBase):
    def __init__(self, spectra_data_id, spectrum_id, id=None, identifications=None, params=None,
                 context=NullMap, **kwargs):
        if identifications is None:
            identifications = []
        self.params = self.prepare_params(params, **kwargs)
        self.identifications = identifications
        self.element = _element(
            "SpectrumIdentificationResult", spectraData_ref=context["SpectraData"][spectra_data_id],
            spectrumID=spectrum_id, id=id)
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            for item in self.identifications:
                item.write(xml_file)
            self.write_params(xml_file)


class IonType(ComponentBase):
    def __init__(self, series, indices, charge_state, measures=None, context=NullMap):
        if measures is None:
            measures = dict()
        self.context = context
        self.series = series
        self.measures = measures
        self.element = _element(
            "IonType", charge=charge_state, index=' '.join([str(int(i)) for i in ensure_iterable(indices)]))

    def write(self, xml_file):
        with self.element(xml_file):
            self.context.param(self.series)(xml_file)
            for measure, values in self.measures.items():
                el = _element(
                    "FragmentArray", measure_ref=self.context["Measure"][measure],
                    values=' '.join([str(float(i)) for i in values]))
                el.write(xml_file)

    all_ion_types = [
        'frag: a ion', 'frag: a ion - H2O', 'frag: a ion - NH3',
        'frag: b ion', 'frag: b ion - H2O', 'frag: b ion - NH3',
        'frag: c ion', 'frag: c ion - H2O', 'frag: c ion - NH3',
        'frag: d ion',
        'frag: immonium ion',
        'frag: internal ya ion',
        'frag: internal yb ion',
        'frag: precursor ion', 'frag: precursor ion - H2O', 'frag: precursor ion - NH3',
        'frag: v ion',
        'frag: w ion',
        'frag: x ion', 'frag: x ion - H2O', 'frag: x ion - NH3',
        'frag: y ion', 'frag: y ion - H2O', 'frag: y ion - NH3',
        'frag: z ion', 'frag: z ion - H2O', 'frag: z ion - NH3',
        'frag: z+1 ion', 'frag: z+2 ion'
    ]

    _ion_pattern = re.compile(
        r"""(?P<prefix>frag:\s?)?
            (?P<key>[abcxyzdvw]|internal [yab]{2}|precursor|immonium|z\+[12])
            (?P<ion>\s?ion\s?)?
            (?=\s?-\s?(?P<loss>NH3|H2O))?""", re.VERBOSE)

    @classmethod
    def guess_ion_type(cls, string):
        match = cls._ion_pattern.search(string)
        if match:
            groups = match.groupdict()
            loss = groups.get("loss")
            if loss is not None:
                loss = " - " + loss
            else:
                loss = ''
            return "frag: {key} ion{loss}".format(key=groups['key'], loss=loss)
        else:
            raise KeyError("No mapping found for %s" % string)


class SpectrumIdentificationItem(ComponentBase):
    def __init__(self, experimental_mass_to_charge,
                 charge_state, peptide_id, peptide_evidence_ids, id, score=None,
                 ion_types=None, params=None, pass_threshold=True, rank=1,
                 calculated_mass_to_charge=None, calculated_pi=None,
                 name=None, mass_table_id=None, sample_id=None,
                 context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.peptide_evidence_refs = [
            context["PeptideEvidence"][peptide_evidence_id]
            for peptide_evidence_id in ensure_iterable(peptide_evidence_ids)]
        self.params = params
        self.score = score
        self.rank = rank
        self.element = _element(
            "SpectrumIdentificationItem", chargeState=charge_state,
            experimentalMassToCharge=experimental_mass_to_charge, id=id, passThreshold=pass_threshold,
            peptide_ref=context['Peptide'][peptide_id], rank=self.rank
        )
        self.element.attrs['calculatedMassToCharge'] = calculated_mass_to_charge
        self.element.attrs['calculatedPI'] = calculated_pi
        if sample_id is not None:
            self.element.attrs['sample_ref'] = context['Sample'][sample_id]
        if mass_table_id is not None:
            self.element.attrs['massTable_ref'] = context['MassTable'][sample_id]
        context['SpectrumIdentificationItem'][id] = self.element.id
        self.context = context
        self.ion_types = self.prepare_ion_types(ion_types)

    def prepare_ion_types(self, ion_types):
        mappings = []
        if ion_types is not None:
            if isinstance(ion_types, (list, tuple)):
                try:
                    if isinstance(ion_types[0], IonType):
                        return ion_types
                except IndexError:
                    return mappings
            elif isinstance(ion_types, Mapping):
                ion_types = list(ion_types.items())
            for series, measures in ion_types:
                measures_ = {}
                for measure, values in measures.items():
                    if measure in ("indices", "charge_state"):
                        continue
                    measures_[measure] = values

                it = IonType(series=IonType.guess_ion_type(series), indices=measures['indices'],
                             charge_state=measures.get('charge_state', 1), measures=measures_,
                             context=self.context)
                mappings.append(it)
        return mappings

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            for peptide_evidence_ref in self.peptide_evidence_refs:
                _element(
                    "PeptideEvidenceRef",
                    peptideEvidence_ref=peptide_evidence_ref).write(
                    xml_file)
            ion_types = ensure_iterable(self.ion_types)
            if self.ion_types is not None and len(ion_types) > 0:
                with element(xml_file, "Fragmentation"):
                    for ion_type in ion_types:
                        ion_type.write(xml_file)
            if isinstance(self.score, CVParam):
                self.score.write(xml_file)
            elif isinstance(self.score, dict):
                self.context.param(self.score)(xml_file)
            else:
                self.context.param(name="score", value=self.score)(xml_file)
            self.write_params(xml_file)


class Measure(ComponentBase):
    def __init__(self, name=None, id=None, param=None, getter=None, context=NullMap):
        if isinstance(name, CVParam):
            param = name
            name = param.name
        elif param is None and name is not None:
            param = context.param(name)
        self.name = name
        self.param = param
        if name is not None and id is None:
            id = name
        self.element = _element("Measure", id=sanitize_id(id), name=name)
        context["Measure"][id] = self.element.id
        self.context = context

        if getter is None:
            getter = operator.itemgetter(name)
        self.getter = getter

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            self.context.param(self.param)(xml_file)

    common_measures = ["product ion %s" % s for s in ("m/z", "intensity", "m/z error")]


class FragmentationTable(ComponentBase):
    def __init__(self, measures=None, context=NullMap):
        if measures is None:
            measures = Measure.common_measures
        self.context = context

        self.element = _element("FragmentationTable")

        self.measures = []
        for measure in measures:
            if isinstance(measure, Measure):
                self.measures.append(measure)
            else:
                measure = Measure(context.param(measure), context=context)
                self.measures.append(measure)

    def write(self, xml_file):
        with self.element(xml_file):
            for measure in self.measures:
                measure.write(xml_file)


class SpectrumIdentificationList(ComponentBase):
    def __init__(self, identification_results, id, fragmentation_table=None, context=NullMap):
        self.identification_results = identification_results
        self.fragmentation_table = fragmentation_table
        self.element = _element("SpectrumIdentificationList", xmlns=_xmlns, id=id)
        context["SpectrumIdentificationList"][id] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            if self.fragmentation_table is not None:
                self.fragmentation_table.write(xml_file)
            for identification_result in self.identification_results:
                identification_result.write(xml_file)


class AnalysisData(GenericCollection):
    def __init__(self, identification_lists=tuple(), protein_detection_lists=tuple(), context=NullMap):
        items = list()
        items.extend(identification_lists)
        items.extend(protein_detection_lists)
        super(AnalysisData, self).__init__("AnalysisData", items, context)


class ProteinDetectionList(ComponentBase):
    def __init__(self, ambiguity_groups=None, count=AUTO, id=1, params=None, context=NullMap, **kwargs):
        self.ambiguity_groups = ambiguity_groups
        self.element = _element("ProteinDetectionList", id=id)
        self.count = count
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.context['ProteinDetectionList'][id] = self.element.id

    def _count_protein_groups(self):
        count = 0
        for pg in self.ambiguity_groups:
            if pg.pass_threshold:
                count += 1
        return count

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for ambiguity_group in self.ambiguity_groups:
                ambiguity_group.write(xml_file)
            if self.count is AUTO:
                count = self._count_protein_groups()
                if count > 0:
                    self.context.param("count of identified proteins", count)(xml_file)
            elif self.count is not None:
                self.context.param("count of identified proteins", self.count)(xml_file)
            self.write_params(xml_file)


class PeptideHypothesis(ComponentBase):
    def __init__(self, peptide_evidence_id, spectrum_identification_ids, params=None,
                 context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.peptide_evidence_id = peptide_evidence_id
        self.spectrum_identification_ids = spectrum_identification_ids
        self.params = params
        self.context = context
        self.element = _element("PeptideHypothesis",
                                peptideEvidence_ref=self.context['PeptideEvidence'][peptide_evidence_id])

    def write(self, xml_file):
        with self.element(xml_file):
            for spectrum_identification_id in self.spectrum_identification_ids:
                el = _element(
                    "SpectrumIdentificationItemRef",
                    spectrumIdentificationItem_ref=self.context[
                        "SpectrumIdentificationItem"][spectrum_identification_id])
                el.write(xml_file)
            self.write_params(xml_file)


class ProteinDetectionHypothesis(ComponentBase):
    def __init__(self, id, db_sequence_id, peptide_hypotheses, pass_threshold=True,
                 leading=True, params=None, name=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.db_sequence_id = db_sequence_id
        self.pass_threshold = pass_threshold
        self.leading = leading
        self.params = params
        self.context = context
        self.peptide_hypotheses = self._coerce_peptide_hypotheses(peptide_hypotheses)
        self.element = _element(
            "ProteinDetectionHypothesis", id=id,
            dBSequence_ref=context["DBSequence"][db_sequence_id],
            passThreshold=pass_threshold)
        if name is not None:
            self.element.attrs['name'] = name
        self.context['ProteinDetectionHypothesis'][id] = self.element.id

    def _coerce_peptide_hypotheses(self, peptides):
        out = []
        for peptide in ensure_iterable(peptides):
            if isinstance(peptide, Mapping):
                peptide = PeptideHypothesis(context=self.context, **peptide)
            out.append(peptide)
        return out

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for peptide_hypothesis in self.peptide_hypotheses:
                peptide_hypothesis.write(xml_file)
            if self.leading is None:
                pass
            elif self.leading:
                self.context.param("leading protein")(xml_file)
            else:
                self.context.param("non-leading protein")(xml_file)
            self.write_params(xml_file)


class ProteinAmbiguityGroup(ComponentBase):
    def __init__(self, id, protein_detection_hypotheses, pass_threshold=True, params=None,
                 context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.protein_detection_hypotheses = protein_detection_hypotheses
        self.params = params
        self.pass_threshold = pass_threshold
        self.context = context
        self.element = _element(
            "ProteinAmbiguityGroup", id=id)
        self.context["ProteinAmbiguityGroup"][id] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for protein in self.protein_detection_hypotheses:
                protein.write(xml_file)
            if self.pass_threshold is not None:
                self.context.param(name="protein group passes threshold",
                                   value=self.pass_threshold)(xml_file)
            self.write_params(xml_file)


# --------------------------------------------------
# Meta-collections


class DataCollection(GenericCollection):
    def __init__(self, inputs, analysis_data, context=NullMap):
        super(DataCollection, self).__init__("DataCollection", [inputs, analysis_data], context)


class SequenceCollection(GenericCollection):
    def __init__(self, db_sequences, peptides, peptide_evidence, context=NullMap):
        super(SequenceCollection, self).__init__("SequenceCollection", chain.from_iterable(
            [db_sequences, peptides, peptide_evidence]))


# --------------------------------------------------
# Software Execution Protocol Information


class Enzyme(ComponentBase):
    def __init__(self, name, missed_cleavages=1, id=None, semi_specific=False, site_regexp=None,
                 min_distance=None, n_term_gain=None, c_term_gain=None, params=None, context=NullMap,
                 **kwargs):
        self.name = name
        if site_regexp is None:
            term = context.term(name)
            try:
                regex_ref = term['has_regexp']
                regex_ent = context.term(regex_ref)
                regex = regex_ent['name']
                site_regexp = regex
            except Exception:
                pass
        self.site_regexp = site_regexp
        self.element = _element(
            "Enzyme", semiSpecific=semi_specific, missedCleavages=missed_cleavages,
            id=id)
        context["Enzyme"][id] = self.element.id
        self.context = context
        self.params = self.prepare_params(params, **kwargs)
        self.min_distance = min_distance
        if min_distance is not None:
            self.element.attrs['minDistance'] = int(min_distance)
        self.n_term_gain = n_term_gain
        if n_term_gain is not None:
            self.element.attrs['nTermGain'] = str(n_term_gain)
        self.c_term_gain = c_term_gain
        if c_term_gain is not None:
            self.element.attrs['cTermGain'] = str(c_term_gain)

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            if self.site_regexp is not None:
                regex = _element("SiteRegexp").element()
                regex.text = etree.CDATA(self.site_regexp)
                xml_file.write(regex)
            with element(xml_file, "EnzymeName"):
                self.context.param(self.name)(xml_file)
            self.write_params(xml_file)


class Enzymes(ComponentBase):
    def __init__(self, enzymes=None, independent=None, context=NullMap):
        self.independent = independent
        self.context = context
        self.enzymes = self._coerce_enzymes(enzymes)
        self.element = _element("Enzymes")
        if self.independent is not None:
            self.element.attrs['independent'] = bool(self.independent)

    def _coerce_enzymes(self, enzymes):
        temp = []
        for enz in ensure_iterable(enzymes):
            if not isinstance(enz, Enzyme):
                enz = Enzyme(context=self.context, **enz)
            temp.append(enz)
        return temp

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            for enzyme in self.enzymes:
                enzyme.write(xml_file)


class _Tolerance(ComponentBase):

    def __init__(self, low, high=None, unit="parts per million", context=NullMap):
        if isinstance(low, NumberBase):
            low = CVParam(
                accession="MS:1001413", ref="PSI-MS", unitCvRef="UO", unitName=unit,
                unitAccession=common_units[unit], value=low,
                name="search tolerance minus value")
        if high is None:
            high = CVParam(
                accession="MS:1001412", ref="PSI-MS", unitCvRef="UO", unitName=unit,
                unitAccession=common_units[unit], value=low.value,
                name="search tolerance plus value")
        elif isinstance(high, NumberBase):
            high = CVParam(
                accession="MS:1001412", ref="PSI-MS", unitCvRef="UO", unitName=unit,
                unitAccession=common_units[unit], value=high,
                name="search tolerance plus value")

        self.low = low
        self.high = high

    def write(self, xml_file):
        with element(xml_file, self.tag_name):
            self.low.write(xml_file)
            self.high.write(xml_file)

    def __iter__(self):
        yield self.low.value
        yield self.high.value


class FragmentTolerance(_Tolerance):
    tag_name = "FragmentTolerance"


class ParentTolerance(_Tolerance):
    tag_name = "ParentTolerance"


class Threshold(ComponentBase):
    no_threshold = CVParam(accession="MS:1001494", ref="PSI-MS", name="no threshold")

    def __init__(self, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(ensure_iterable(params), **kwargs)
        self.params = params
        self.context = context

    def write(self, xml_file):
        with element(xml_file, "Threshold"):
            if not self.params:
                self.no_threshold(xml_file)
            else:
                self.write_params(xml_file)


class SpectrumIdentificationProtocol(ComponentBase):
    def __init__(self, search_type, analysis_software_id=1, id=1, additional_search_params=tuple(),
                 modification_params=tuple(), enzymes=tuple(), fragment_tolerance=None, parent_tolerance=None,
                 threshold=None, filters=None, context=NullMap):
        self.context = context
        if threshold is None:
            threshold = Threshold(context=context)
        elif not isinstance(threshold, Threshold):
            threshold = Threshold(threshold, context=context)
        if not isinstance(parent_tolerance, ParentTolerance):
            if isinstance(parent_tolerance, NumberBase):
                parent_tolerance = ParentTolerance(parent_tolerance)
            elif isinstance(parent_tolerance, (tuple, list)):
                parent_tolerance = ParentTolerance(*parent_tolerance)
            elif parent_tolerance is None:
                pass
            else:
                raise ValueError("Cannot infer ParentTolerance from %r" % (parent_tolerance,))
        if not isinstance(fragment_tolerance, FragmentTolerance):
            if isinstance(fragment_tolerance, NumberBase):
                fragment_tolerance = FragmentTolerance(fragment_tolerance)
            elif isinstance(fragment_tolerance, (tuple, list)):
                fragment_tolerance = FragmentTolerance(*fragment_tolerance)
            elif fragment_tolerance is None:
                pass
            else:
                raise ValueError("Cannot infer FragmentTolerance from %r" % (fragment_tolerance,))
        if not isinstance(enzymes, Enzymes):
            enzymes = Enzymes(ensure_iterable(enzymes), context=context)
        self.parent_tolerance = parent_tolerance
        self.fragment_tolerance = fragment_tolerance
        self.threshold = threshold
        self.enzymes = enzymes
        temp = []
        for mod in ensure_iterable(modification_params):
            if isinstance(mod, SearchModification):
                temp.append(mod)
            else:
                temp.append(
                    SearchModification(
                        context=context, **mod))
        modification_params = temp
        self.modification_params = modification_params
        self.additional_search_params = self.prepare_params(additional_search_params)
        self.search_type = search_type
        temp = []
        for filt in ensure_iterable(filters):
            if isinstance(filt, Filter):
                temp.append(filt)
            else:
                temp.append(Filter(context=context, **filt))
        filters = temp
        self.filters = filters
        self.element = _element(
            "SpectrumIdentificationProtocol", id=id,
            analysisSoftware_ref=context['AnalysisSoftware'][analysis_software_id])
        self.context["SpectrumIdentificationProtocol"][id] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            with element(xml_file, "SearchType"):
                self.context.param(self.search_type)(xml_file)
            if self.additional_search_params:
                with element(xml_file, "AdditionalSearchParams"):
                    self.write_params(xml_file, self.additional_search_params)
            with element(xml_file, "ModificationParams"):
                for mod in self.modification_params:
                    mod.write(xml_file)
            self.enzymes.write(xml_file)
            if self.fragment_tolerance is not None:
                self.fragment_tolerance.write(xml_file)
            if self.parent_tolerance is not None:
                self.parent_tolerance.write(xml_file)
            self.threshold.write(xml_file)
            if self.filters:
                with element(xml_file, "DatabaseFilters"):
                    for filt in self.filters:
                        filt.write(xml_file)


class Filter(ComponentBase):
    def __init__(self, filter_type, include=None, exclude=None, context=NullMap):
        self.filter_type = filter_type
        self.include = include
        self.exclude = exclude
        self.context = context
        self.element = _element("Filter")

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            with element(xml_file, "FilterType"):
                self.context.param(self.filter_type)(xml_file)
            if self.include:
                with element(xml_file, "Include"):
                    self.write_params(xml_file, self.prepare_params(self.include))
            if self.exclude:
                with element(xml_file, "Exclude"):
                    self.write_params(xml_file, self.prepare_params(self.exclude))


class SpecificityRules(ParameterContainer):
    def __init__(self, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(SpecificityRules, self).__init__("SpecificityRules", params, context=context)


class SearchModification(ComponentBase):
    def __init__(self, mass_delta, fixed, residues, specificity=None,
                 params=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        if specificity is not None:
            if not isinstance(specificity, (tuple, list)):
                specificity = [specificity]
            if not isinstance(specificity[0], SpecificityRules):
                specificity = [
                    SpecificityRules(s, context=context) for s in specificity]
        params.extend(kwargs.items())
        self.params = params
        self.mass_delta = mass_delta
        self.fixed = fixed
        self.residues = ''.join(residues)
        self.specificity = specificity
        self.context = context
        self.element = _element(
            "SearchModification", fixedMod=fixed, massDelta=mass_delta,
            residues=self.residues)

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            self.write_params(xml_file)
            if self.specificity is not None:
                with _element("SpecificityRules"):
                    self.write_params(xml_file, self.specificity)


class ProteinDetectionProtocol(ComponentBase):
    def __init__(self, id=1, params=None, analysis_software_id=1, threshold=None,
                 context=NullMap, **kwargs):
        if threshold is None:
            threshold = Threshold(context=context)
        elif isinstance(threshold, (list, tuple)):
            threshold = Threshold(threshold, context=context)
        params = self.prepare_params(params, **kwargs)
        self.context = context
        self.threshold = threshold
        self.params = params
        self.analysis_software_id = analysis_software_id
        self.element = _element(
            "ProteinDetectionProtocol", id=id,
            analysisSoftware_ref=context["AnalysisSoftware"][analysis_software_id])
        self.context["ProteinDetectionProtocol"][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            self.threshold.write(xml_file)
            if self.params:
                with element(xml_file, "AnalysisParams"):
                    self.write_params(xml_file)


class AnalysisProtocolCollection(GenericCollection):
    def __init__(self, spectrum_identification_protocols=tuple(),
                 protein_detection_protocols=tuple(), context=NullMap):
        items = list()
        items.extend(spectrum_identification_protocols)
        items.extend(protein_detection_protocols)
        super(AnalysisProtocolCollection, self).__init__(self, items, context)


# --------------------------------------------------
# Analysis Collection - Data-to-Analysis

class SpectrumIdentification(ComponentBase):
    def __init__(self, spectra_data_ids_used=None, search_database_ids_used=None, spectrum_identification_list_id=1,
                 spectrum_identification_protocol_id=1, id=1, context=NullMap):
        self.spectra_data_ids_used = [context["SpectraData"][x] for x in (spectra_data_ids_used or [])]
        self.search_database_ids_used = [context["SearchDatabase"][x] for x in (search_database_ids_used or [])]

        self.element = _element(
            "SpectrumIdentification", id=id,
            spectrumIdentificationList_ref=context["SpectrumIdentificationList"][
                spectrum_identification_list_id],
            spectrumIdentificationProtocol_ref=context["SpectrumIdentificationProtocol"][
                spectrum_identification_protocol_id])
        self.context = context
        context["SpectrumIdentification"] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for spectra_data_id in self.spectra_data_ids_used:
                _element("InputSpectra", spectraData_ref=spectra_data_id).write(xml_file)
            for search_database_id in self.search_database_ids_used:
                _element("SearchDatabaseRef", searchDatabase_ref=search_database_id).write(xml_file)


class ProteinDetection(ComponentBase):
    def __init__(self, spectrum_identification_ids_used, protein_detection_list_id=1,
                 protein_detection_protocol_id=1, id=1, context=NullMap):
        self.spectrum_identification_ids_used = [context["SpectrumIdentificationList"][x]
                                                 for x in (spectrum_identification_ids_used or [])]
        self.protein_detection_list_id = protein_detection_list_id
        self.protein_detection_protocol_id = protein_detection_protocol_id
        self.context = context
        self.element = _element(
            "ProteinDetection", id=id,
            proteinDetectionProtocol_ref=context['ProteinDetectionProtocol'][protein_detection_protocol_id],
            proteinDetectionList_ref=context['ProteinDetectionList'][protein_detection_list_id])

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for sid in self.spectrum_identification_ids_used:
                _element("InputSpectrumIdentifications",
                         spectrumIdentificationList_ref=sid).write(xml_file)


# --------------------------------------------------
# Misc. Providence Management


DEFAULT_CONTACT_ID = "PERSON_DOC_OWNER"
DEFAULT_ORGANIZATION_ID = "ORG_DOC_OWNER"


class CVList(ComponentBase):
    def __init__(self, cv_list=None, context=NullMap):
        if cv_list is None:
            cv_list = default_cv_list
        self.cv_list = cv_list

    def write(self, xml_file):
        with element(xml_file, 'cvList'):
            for member in self.cv_list:
                tag = _element(
                    "cv", id=member.id, fullName=member.full_name,
                    uri=member.uri)
                if member.version is not None:
                    tag.attrs['version'] = member.version
                if member.options:
                    tag.attrs.update(member.options)
                xml_file.write(tag.element(with_id=True))

    def __iter__(self):
        return iter(self.cv_list)


class AnalysisSoftware(ComponentBase):
    def __init__(self, name, id=1, version=None, uri=None, contact=DEFAULT_CONTACT_ID,
                 role='software vendor', customization=None, context=NullMap, **kwargs):
        self.name = name
        self.version = version
        self.uri = uri
        self.contact = contact
        self.role = role
        self.kwargs = kwargs
        self.element = _element("AnalysisSoftware", id=id, name=self.name, version=self.version, uri=self.uri)
        self.context = context
        self.context["AnalysisSoftware"][id] = self.element.id
        self.customization = customization

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            with element(xml_file, "ContactRole", contact_ref=self.contact):
                with element(xml_file, "Role"):
                    self.write_params(xml_file, self.prepare_params(self.role))
            with element(xml_file, "SoftwareName"):
                name_param = self.context.param(name=self.name)
                # if the software name is not a controlled vocabulary term, specify it using
                # the standard "unreleased custom software tool" term-value pair
                if isinstance(name_param, UserParam):
                    name_param = self.context.param("MS:1000799", value=self.name)
                name_param(xml_file)
            if self.customization is not None:
                with element(xml_file, "Customizations"):
                    xml_file.write('\n'.join(ensure_iterable(self.customization)))


class Provider(ComponentBase):
    def __init__(self, id="PROVIDER", role='researcher', contact=DEFAULT_CONTACT_ID, context=NullMap):
        self.id = id
        self.contact = contact
        self.role = role
        self.context = context
        self.element = _element('Provider', id=id, xmlns=_xmlns)
        self.context['Provider'][id] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            with element(xml_file, "ContactRole", contact_ref=self.contact):
                with element(xml_file, "Role"):
                    self.write_params(xml_file, self.prepare_params(self.role))


class Person(ComponentBase):
    def __init__(self, first_name=None, last_name=None, middle_initial=None, id=DEFAULT_CONTACT_ID,
                 affiliation=DEFAULT_ORGANIZATION_ID, params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.first_name = first_name
        self.middle_initial = middle_initial
        self.last_name = last_name
        self.id = id
        self.element = _element(
            "Person", firstName=first_name, lastName=last_name,
            midInitials=middle_initial, id=id)
        self.context = context
        self.context["Person"][id] = self.element.id
        self.affiliation = self.context["Organization"][affiliation]

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            if self.affiliation is not None:
                _element('Affiliation', organization_ref=self.affiliation).write(xml_file)
            self.write_params(xml_file)


class Organization(ComponentBase):
    def __init__(self, name=None, id=DEFAULT_ORGANIZATION_ID, params=None,
                 parent=None, context=NullMap, **kwargs):
        self.name = name
        self.id = id
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element("Organization", name=name, id=id)
        self.context = context
        self.context["Organization"][id] = self.element.id
        self.parent = self.context["Organization"][parent]

    def write(self, xml_file):
        xml_file.write(self.element.element())
        self.write_params(xml_file)
        if self.parent is not None:
            _element("Parent", organization_ref=self.parent).write(xml_file)


DEFAULT_PERSON = Person()
DEFAULT_ORGANIZATION = Organization()


class AuditCollection(ComponentBase):
    def __init__(self, persons=None, organizations=None, context=NullMap):
        if persons is None:
            persons = (DEFAULT_PERSON,)
        if organizations is None:
            organizations = (DEFAULT_ORGANIZATION,)
        self.persons = persons
        self.organizations = organizations

    def write(self, xml_file):
        with element(xml_file, "AuditCollection", xmlns=_xmlns):
            for person in self.persons:
                person.write(xml_file)
            for organization in self.organizations:
                organization.write(xml_file)


class AnalysisSampleCollection(ComponentBase):
    def __init__(self, samples=None, context=NullMap):
        self.samples = ensure_iterable(samples)
        self.context = context
        self.element = _element("AnalysisSampleCollection")

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            for sample in self.samples:
                sample.write(xml_file)


class Sample(ComponentBase):
    def __init__(self, id, name=None, contact=None, role=None, sub_samples=None,
                 params=None, context=NullMap, **kwargs):
        self.id = id
        self.name = name
        self.contact = contact
        self.role = ensure_iterable(role)
        self.sub_samples = ensure_iterable(sub_samples)
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.element = _element("Sample", id=self.id, name=self.name)
        self.context['Sample'][self.id] = self.element.id

    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            with element(xml_file, "ContactRole", contact_ref=self.contact):
                with element(xml_file, "Role"):
                    self.write_params(xml_file, self.prepare_params(self.role))
            for sample_ref in self.sub_samples:
                _element(
                    "SubSample",
                    sample_ref=self.context['Sample'][sample_ref]).write(xml_file)
            self.write_params(xml_file)



import warnings
from collections import defaultdict, OrderedDict
from functools import partial, update_wrapper
from contextlib import contextmanager

from .utils import add_metaclass, ensure_iterable, Mapping

from .xml import (
    id_maker, CVParam, UserParam,
    ParamGroupReference, _element)


class ChildTrackingMeta(type):
    def __new__(cls, name, parents, attrs):
        if not hasattr(cls, "_cache"):
            cls._cache = defaultdict(dict)
        new_type = type.__new__(cls, name, parents, attrs)
        ns = getattr(new_type, "component_namespace", None)
        cls._cache[ns][name] = new_type
        return new_type

    @classmethod
    def resolve_component(self, namespace, name):
        try:
            return self._cache[namespace][name]
        except KeyError:
            return self._cache[None][name]


class SpecializedContextCache(OrderedDict):
    def __init__(self, type_name):
        super(SpecializedContextCache, self).__init__()
        self.type_name = type_name
        self.bijection = dict()

    def __getitem__(self, key):
        try:
            item = dict.__getitem__(self, key)
            return item
        except KeyError:
            if key is None:
                return None
            warnings.warn("No reference was found for %r in %s" % (key, self.type_name), stacklevel=3)
            new_value = id_maker(self.type_name, key)
            self[key] = new_value
            return new_value

    def __setitem__(self, key, value):
        super(SpecializedContextCache, self).__setitem__(key, value)
        self.bijection[value] = key

    def __repr__(self):
        return '%s\n%s' % (self.type_name, dict.__repr__(self))


class VocabularyResolver(object):
    def __init__(self, vocabularies=None):
        self.vocabularies = vocabularies

    def get_vocabulary(self, id):
        for vocab in self.vocabularies:
            if vocab.id == id:
                return vocab
            try:
                if vocab.full_name == id:
                    return vocab
            except AttributeError:
                pass
        raise KeyError(id)

    def param_group_reference(self, id):
        return ParamGroupReference(id)

    def param(self, name, value=None, cv_ref=None, **kwargs):
        accession = kwargs.get("accession")

        if isinstance(name, CVParam):
            return name
        elif isinstance(name, ParamGroupReference):
            return name
        elif isinstance(name, (tuple, list)) and value is None:
            name, value = name
        elif isinstance(name, Mapping):
            mapping = dict(name)
            if len(mapping) == 1 and 'ref' in mapping:
                return self.param_group_reference(mapping['ref'])
            value = value or mapping.pop('value', None)
            accession = accession or mapping.pop("accession", None)
            cv_ref = cv_ref or mapping.pop("cv_ref", None) or mapping.pop("cvRef", None)
            unit_name = mapping.pop("unit_name", None) or mapping.pop("unitName", None)
            unit_accession = mapping.pop("unit_accession", None) or mapping.pop("unitAccession", None)
            unit_cv_ref = mapping.pop('unit_cv_ref', None) or mapping.pop('unitCvRef', None)
            name = mapping.pop('name', None)
            if name is None:
                if len(mapping) == 1:
                    name, value = tuple(mapping.items())[0]
                else:
                    raise ValueError("Could not coerce parameter from %r" % (mapping,))
            else:
                kwargs.update({k: v for k, v in mapping.items()
                               if k not in (
                    "name", "value", "accession")})
                # case normalize unit information so that cvParam can detect them
                if unit_name is not None:
                    kwargs.setdefault("unit_name", unit_name)
                if unit_accession is not None:
                    kwargs.setdefault("unit_accession", unit_accession)
                if unit_cv_ref is not None:
                    kwargs.setdefault("unit_cv_ref", unit_cv_ref)
        unit_name = kwargs.get("unit_name")
        unit_accession = kwargs.get("unit_accession")
        unit_ref = kwargs.get('unit_cv_ref')
        if unit_name is not None or unit_accession is not None:
            if unit_accession is not None:
                unit_term, source = self.term(unit_accession, include_source=True)
                unit_name = unit_term.name
                unit_accession = unit_term.id
                unit_ref = source.id
            elif unit_name is not None:
                unit_term, source = self.term(unit_name, include_source=True)
                unit_name = unit_term.name
                unit_accession = unit_term.id
                unit_ref = source.id
            kwargs['unit_name'] = unit_name
            kwargs['unit_accession'] = unit_accession
            kwargs['unit_cv_ref'] = unit_ref

        if name is None:
            raise ValueError("Could not coerce parameter from %r, %r, %r" % (name, value, kwargs))
        if cv_ref is None:
            for cv in self.vocabularies:
                try:
                    term = cv[name]
                    name = term["name"]
                    accession = term["id"]
                    cv_ref = cv.id
                except KeyError:
                    pass
        if cv_ref is None:
            return UserParam(name=name, value=value, **kwargs)
        else:
            kwargs.setdefault("ref", cv_ref)
            kwargs.setdefault("accession", accession)
            return CVParam(name=name, value=value, **kwargs)

    def term(self, name, include_source=False):
        deferred = None
        for cv in self.vocabularies:
            try:
                term = cv[name]
                if term.get("is_obsolete", False):
                    deferred = term, cv
                    raise KeyError(name)
                if include_source:
                    return term, cv
                else:
                    return term
            except KeyError:
                pass
        else:
            if deferred:
                if include_source:
                    return deferred
                else:
                    return deferred[0]
            raise KeyError(name)

    def load_vocabularies(self):
        for vocab in self.vocabularies:
            vocab.load()

    def prepare_params(self, params):
        out = []
        for param in ensure_iterable(params):
            out.append(self.param(param))
        return out


class DocumentContext(dict, VocabularyResolver):
    def __init__(self, vocabularies=None):
        dict.__init__(self)
        VocabularyResolver.__init__(self, vocabularies)

    def param_group_reference(self, id):
        # This is a inelegant, as ReferenceableParamGroup is not part document type
        # independent, and may not be consistent
        param_group_index = self['ReferenceableParamGroup']
        return ParamGroupReference(param_group_index[id])

    def __getitem__(self, key):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        dict.__setitem__(self, key, value)

    def __missing__(self, key):
        if not isinstance(key, str):
            if isinstance(key, (type, ReprBorrowingPartial)):
                key = key.__name__
        self[key] = SpecializedContextCache(key)
        return self[key]


NullMap = DocumentContext()


class ReprBorrowingPartial(partial):
    """
    Create a partial instance that uses the wrapped callable's
    `__repr__` method instead of a generic partial
    """
    def __init__(self, func, *args, **kwargs):
        self._func = func
        # super(ReprBorrowingPartial, self).__init__(func, *args, **kwargs)
        update_wrapper(self, func)

    @property
    def type(self):
        return self._func

    def __repr__(self):
        return repr(self.func)

    def __getattr__(self, name):
        return getattr(self._func, name)

    def ensure(self, data):
        if not isinstance(data, self.type):
            return self(**data)
        else:
            if data.context is not self.context:
                raise ValueError("Cannot bind a component from another context")
        return data


class ComponentDispatcherBase(object):
    """
    A container for a :class:`DocumentContext` which provides
    an automatically parameterized version of all :class:`ComponentBase`
    types which use this instance's context.

    Attributes
    ----------
    context : :class:`DocumentContext`
        The mapping responsible for managing the global
        state of all created components.
    """
    def __init__(self, context=None, vocabularies=None, component_namespace=None):
        if vocabularies is None:
            vocabularies = []
        if context is None:
            context = DocumentContext(vocabularies=vocabularies)
        else:
            if vocabularies is not None:
                context.vocabularies.extend(vocabularies)
        self.component_namespace = component_namespace
        self.context = context

    def __getattr__(self, name):
        """
        Provide access to an automatically parameterized
        version of all :class:`ComponentBase` types which
        use this instance's context.

        Parameters
        ----------
        name : str
            Component Name

        Returns
        -------
        ReprBorrowingPartial
            A partially parameterized instance constructor for
            the :class:`ComponentBase` type requested.
        """
        component = ChildTrackingMeta.resolve_component(self.component_namespace, name)
        tp = ReprBorrowingPartial(component, context=self.context)
        tp.context = self.context
        return tp

    def ensure_component(self, data, tp):
        if isinstance(data, tp.type):
            if self.context is not data.context:
                raise ValueError("Cannot bind a component from another context")
            return data
        else:
            return tp(**data)

    def register(self, entity_type, id):
        """
        Pre-declare an entity in the document context. Ensures that
        a reference look up will be satisfied.

        Parameters
        ----------
        entity_type : str
            An entity type, either a tag name or a component name
        id : int
            The unique id number for the thing registered

        Returns
        -------
        str
            The constructed reference id
        """
        if isinstance(id, int):
            value = id_maker(entity_type, id)
        else:
            value = str(id)
        self.context[entity_type][id] = value
        return value

    @property
    def vocabularies(self):
        return self.context.vocabularies

    def load_vocabularies(self):
        self.context.load_vocabularies()

    def param(self, *args, **kwargs):
        return self.context.param(*args, **kwargs)

    def prepare_params(self, params):
        return self.context.prepare_params(params)

    def term(self, *args, **kwargs):
        return self.context.term(*args, **kwargs)

    def get_vocabulary(self, *args, **kwargs):
        return self.context.get_vocabulary(*args, **kwargs)

# ------------------------------------------
# Base Component Definitions


@add_metaclass(ChildTrackingMeta)
class ComponentBase(object):
    """A base class for all parts of an XML document which
    describe structures composed of more than a single XML
    tag without any children. In addition to wrapping additional
    descriptive data, this type's metaclass is :class:`ChildTrackingMeta`
    which enables :class:`ComponentDispatcherBase` to automatically bind
    a :class:`DocumentContext` object to the `context` paramter of the
    constructor of dynamically generated wrappers.

    Forwards any missing attribute requests to :attr:`element` for resolution
    against's the XML tag's attributes.
    """
    is_open = False
    _entering = None

    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, key):
        if key == "element":
            raise AttributeError(
                ("The `element` attribute failed to instantiate "
                 "or is being accessed too early."))
        try:
            return self.element.attrs[key]
        except KeyError:
            raise AttributeError(key)

    def write(self, xml_file):
        raise NotImplementedError()

    @contextmanager
    def begin(self, xml_file=None, with_id=True):
        if xml_file is None:
            xml_file = getattr(self, "xml_file", None)
            if xml_file is None:
                raise ValueError("xml_file must be provided if this component is not bound!")
        self._entering = self.element(xml_file, with_id=with_id).__enter__()
        self.is_open = True
        yield

    def __call__(self, xml_file):
        self.write(xml_file)

    def __repr__(self):
        return "%s\n%s" % (
            self.element, "\n".join([
                "  %s: %r" % (k, v) for k, v in self.__dict__.items()
                if k not in ("context", "element") and not k.startswith("_")])
        )

    def prepare_params(self, params, **kwargs):
        if isinstance(params, Mapping):
            if ("name" not in params and "accession" not in params) and ("ref" not in params):
                params = list(params.items())
            else:
                params = [params]
        elif isinstance(params, (list, tuple)):
            params = list(params)
        else:
            params = list(ensure_iterable(params)) or []
        params.extend(kwargs.items())
        return params

    def add_param(self, param):
        if isinstance(param, list):
            self.params.extend(param)
        else:
            self.params.append(param)
        return self

    def write_params(self, xml_file, params=None):
        if params is None:
            params = self.params
        params = self.prepare_params(params)
        user_params = []
        cv_params = []
        references = []
        for param in params:
            param = self.context.param(param)
            if isinstance(param, ParamGroupReference):
                references.append(param)
            elif isinstance(param, UserParam):
                user_params.append(param)
            else:
                cv_params.append(param)
        for param in (references + cv_params + user_params):
            param(xml_file)


class ParameterContainer(ComponentBase):
    """An base class for a component whose only purpose
    is to contain one or more cv- or userParams.

    Attributes
    ----------
    context : DocumentContext
        The document metadata store
    element : lxml.etree.Element
        The XML tag object to be written
    params : list
        The list of parameters to include
    """
    def __init__(self, tag_name, params=None, element_args=None, context=NullMap, **kwargs):
        if element_args is None:
            element_args = dict()
        if params is None:
            params = []
        self.params = self.prepare_params(params, **kwargs)
        self.context = context
        self.element = _element(tag_name, **element_args)

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            for param in self.params:
                self.context.param(param)(xml_file)


class IDParameterContainer(ParameterContainer):
    def write(self, xml_file):
        with self.element(xml_file, with_id=True):
            for param in self.params:
                self.context.param(param)(xml_file)



from lxml import etree

from psims.controlled_vocabulary import load_psims
from psims.utils import simple_repr


MUST = "MUST"
MAY = "MAY"
AND = "AND"
OR = "OR"


static_vocabularies = {
    "MS": load_psims(),
}


class RuleParser(object):
    def __init__(self, rules, vocabularies=None):
        if not vocabularies:
            vocabularies = static_vocabularies.copy()
        self.nsmap = {'ms': 'http://psi.hupo.org/ms/mzml'}
        self.vocabularies = vocabularies
        self.rules = list(rules) or []

    @classmethod
    def from_file(cls, path):
        tree = etree.parse(path)
        rules = map(CVRule.from_element, tree.findall(".//CvMappingRule"))
        return cls(rules)

    def get_rule_targets(self, rule, document):
        return document.xpath(rule.cv_element_path, namespaces=self.nsmap)

    def test_rule(self, rule, document):
        targets = self.get_rule_targets(rule, document)
        term_satisfied = []
        for term in rule.terms:
            cv = self.vocabularies[term.cv_identifier]
            satisfied = []
            for match in targets:
                if self._is_instance_of(match, cv[term.term_accession], cv.allow_children):
                    satisfied.append(match)


    def _is_instance_of(self, term, accession, allow_children=True):
        if not allow_children:
            return term.id == accession
        while term is not None:
            if term.id == accession:
                return True
            term = term.parent()
        return False


def _fix_element_path(elpath, nsprefix):
    steps = elpath.split("/")
    acc = ['/']
    for step in steps:
        if not step:
            continue
        if not step.startswith("@"):
            acc.append("%s:%s" % (nsprefix, step))
        else:
            acc.append(step)
    return '/'.join(acc)


class CVRule(object):
    def __init__(self, id, scope_path, cv_element_path, requirement_level, combinator, terms):
        self.id = id
        self.scope_path = scope_path
        self.cv_element_path = cv_element_path
        self.requirement_level = requirement_level
        self.combinator = combinator
        self.terms = terms

    __repr__ = simple_repr

    @classmethod
    def from_element(cls, element):
        terms = []
        for term in element.getchildren():
            if isinstance(term, etree._Comment):
                continue
            terms.append(CVTerm.from_element(term))
        attrs = element.attrib
        inst = cls(
            attrs.get('id'),
            _fix_element_path(attrs.get('scopePath'), 'ms'),
            _fix_element_path(attrs.get('cvElementPath'), 'ms'),
            attrs.get('requirementLevel'),
            attrs.get('cvTermsCombinatorLogic'),
            terms)
        return inst


def parsebool(text):
    return {
        "true": True,
        "false": False,
        None: None
    }[text]


class CVTerm(object):
    def __init__(self, term_accession, use_term_name, use_term, term_name, is_repeatable, allow_children,
                 cv_identifier):
        self.term_accession = term_accession
        self.use_term_name = use_term_name
        self.use_term = use_term
        self.term_name = term_name
        self.is_repeatable = is_repeatable
        self.allow_children = allow_children
        self.cv_identifier = cv_identifier

    __repr__ = simple_repr

    @classmethod
    def from_element(cls, element):
        attrs = element.attrib
        inst = cls(
            attrs.get("termAccession"),
            parsebool(attrs.get('useTermName')),
            parsebool(attrs.get('useTerm')),
            attrs.get("termName"),
            parsebool(attrs.get("isRepeatable")),
            parsebool(attrs.get("allowChildren")),
            attrs.get("cvIdentifierRef"))
        return inst



from .controlled_vocabulary import (
    ControlledVocabulary, obo_cache, OBOCache, load_psims)


from . import unimod
from .unimod import load as load_unimod

from .obo import (
    OBOParser)

from .entity import Entity
from .reference import Reference
from .relationship import Relationship


__all__ = [
    "ControlledVocabulary", "obo_cache", "OBOCache", "OBOParser",
    "obo_cache", "load_psims", "unimod", "load_unimod",
    "Entity", "Reference", "Relationship"
]



from .controlled_vocabulary import (
    ControlledVocabulary, OBOParser,
    load_psims, load_unimod,
    obo_cache, OBOCache)

from .mzml import (
    MzMLWriter, ARRAY_TYPES, compression_map,
    MZ_ARRAY, INTENSITY_ARRAY, CHARGE_ARRAY,
    components as mzml_components,
    default_cv_list as default_mzml_cv_list)

from .mzid import (
    MzIdentMLWriter, default_cv_list as default_mzid_cv_list,
    components as mzid_components,)

from .version import version as __version__



import numbers
import warnings

from io import BytesIO
from collections import defaultdict, OrderedDict

try:
    from collections import Sequence, Mapping
except ImportError:
    from collections.abc import Sequence, Mapping

import numpy as np

from psims.xml import XMLWriterMixin, XMLDocumentWriter
from psims import compression

from . import components
from .components import (
    ComponentDispatcher, element,
    default_cv_list, MzML, InstrumentConfiguration)

from .binary_encoding import (
    encode_array, COMPRESSION_NONE, COMPRESSION_ZLIB,
    encoding_map, compression_map, dtype_to_encoding)

from .utils import ensure_iterable

from .index import MzMLIndexer


MZ_ARRAY = 'm/z array'
INTENSITY_ARRAY = 'intensity array'
CHARGE_ARRAY = 'charge array'
TIME_ARRAY = "time array"
NON_STANDARD_ARRAY = 'non-standard data array'

ARRAY_TYPES = [
    'm/z array',
    'intensity array',
    'charge array',
    'signal to noise array',
    'time array',
    'wavelength array',
    'flow rate array',
    'pressure array',
    'temperature array',
    'mean drift time array',
    'mean charge array',
    'resolution array',
    'baseline array'
]


class DocumentSection(ComponentDispatcher, XMLWriterMixin):

    def __init__(self, section, writer, parent_context, section_args=None, **kwargs):
        if section_args is None:
            section_args = dict()
        section_args.update(kwargs)
        super(DocumentSection, self).__init__(parent_context)
        self.section = section
        self.writer = writer
        self.section_args = section_args

    def __enter__(self):
        self.toplevel = element(self.writer, self.section, **self.section_args)
        self.toplevel.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        self.toplevel.__exit__(exc_type, exc_value, traceback)
        self.writer.flush()


class SpectrumListSection(DocumentSection):
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(SpectrumListSection, self).__init__(
            "spectrumList", writer, parent_context, section_args=section_args,
            **kwargs)
        self.section_args.setdefault("count", 0)
        data_processing_method = self.section_args.pop(
            "data_processing_method", None)
        try:
            self.section_args["defaultDataProcessingRef"] = self.context[
                "DataProcessing"][data_processing_method]
        except KeyError:
            try:
                self.section_args["defaultDataProcessingRef"] = list(
                    self.context["DataProcessing"].values())[0]
            except IndexError:
                warnings.warn("No Data Processing method found. mzML file may not be fully standard-compliant")


class ChromatogramListSection(DocumentSection):
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(ChromatogramListSection, self).__init__(
            "chromatogramList", writer, parent_context,
            section_args=section_args, **kwargs)
        self.section_args.setdefault("count", 0)
        data_processing_method = self.section_args.pop(
            "data_processing_method", None)
        try:
            self.section_args["defaultDataProcessingRef"] = self.context[
                "DataProcessing"][data_processing_method]
        except KeyError:
            try:
                self.section_args["defaultDataProcessingRef"] = list(
                    self.context["DataProcessing"].values())[0]
            except IndexError:
                warnings.warn("No Data Processing method found. mzML file may not be fully standard-compliant")


class RunSection(DocumentSection):
    """Describes a `<run>` tag. Implemented as a section to provide a more
    expressive API
    """
    def __init__(self, writer, parent_context, section_args=None, **kwargs):
        super(RunSection, self).__init__(
            "run", writer, parent_context, section_args=section_args, **kwargs)
        instrument_configuration_name = self.section_args.pop(
            "instrument_configuration", None)
        if instrument_configuration_name is not None:
            self.section_args["defaultInstrumentConfigurationRef"] = self.context[
                "InstrumentConfiguration"][instrument_configuration_name]
        source_file_name = self.section_args.pop("source_file", None)
        if source_file_name is not None:
            self.section_args["defaultSourceFileRef"] = self.context[
                "SourceFile"][source_file_name]
        sample_id = self.section_args.pop("sample", None)
        if sample_id is not None:
            self.section_args["sampleRef"] = self.context['Sample'][sample_id]


class MzMLWriter(ComponentDispatcher, XMLDocumentWriter):
    """A high level API for generating mzML XML files from simple Python objects.

    This class depends heavily on lxml's incremental file writing API which in turn
    depends heavily on context managers. Almost all logic is handled inside a context
    manager and in the context of a particular document. Since all operations assume
    that they have access to a universal identity map for each element in the document,
    that map is centralized in this class.

    MzMLWriter inherits from :class:`.ComponentDispatcher`, giving it a :attr:`context`
    attribute and access to all `Component` objects pre-bound to that context with attribute-access
    notation.

    Attributes
    ----------
    chromatogram_count : int
        A count of the number of chromatograms written
    spectrum_count : int
        A count of the number of spectrums written
    """
    toplevel_tag = MzML

    def __init__(self, outfile, close=False, vocabularies=None, **kwargs):
        if vocabularies is None:
            vocabularies = []
        vocabularies = list(default_cv_list) + list(vocabularies)
        ComponentDispatcher.__init__(self, vocabularies=vocabularies)
        XMLDocumentWriter.__init__(self, outfile, close, **kwargs)
        self.spectrum_count = 0
        self.chromatogram_count = 0
        self.default_instrument_configuration = None

    def software_list(self, software_list):
        n = len(software_list)
        if n:
            software_list = [self.Software.ensure(sw) for sw in ensure_iterable(software_list)]
        self.SoftwareList(software_list).write(self)

    def file_description(self, file_contents=None, source_files=None, params=None):
        """Writes the `<fileDescription>` section of the document using the
        provided parameters.

        From the specification:
        .. note::
            Information pertaining to the entire mzML file (i.e. not specific
            to any part of the data set) is stored here.

        Parameters
        ----------
        file_contents : list, optional
            A list or other iterable of str, dict, or *Param-types which will
            be placed in the `<fileContent>` element.
        source_files : list, optional
            A list or other iterable of dict or :class:`.SourceFile`-like objects
            to be placed in the `<sourceFileList>` element
        """
        fd = self.FileDescription(file_contents, [self.SourceFile(**sf) for sf in ensure_iterable(source_files)])
        fd.write(self.writer)

    def instrument_configuration_list(self, instrument_configurations=None):
        configs = [
            self.InstrumentConfiguration.ensure(ic) if not isinstance(
                ic, InstrumentConfiguration) else ic
            for ic in ensure_iterable(
                instrument_configurations)]
        self.InstrumentConfigurationList(configs).write(self)

    def data_processing_list(self, processing_methods=None):
        methods = [
            self.DataProcessing.ensure(dp) for dp in ensure_iterable(processing_methods)]
        self.DataProcessingList(methods).write(self)

    def reference_param_group_list(self, groups=None):
        groups = [
            self.ReferenceableParamGroup.ensure(g) for g in ensure_iterable(groups)]
        self.ReferenceableParamGroupList(groups).write(self)

    def sample_list(self, samples):
        for i, sample in enumerate(samples):
            sample_id = sample.get('id')
            sample_name = sample.get("name")

            if sample_id is None and sample_name is not None:
                sample_id = "%s_%d_id" % (sample_name, i)
            elif sample_id is not None and sample_name is None:
                sample_name = str(sample_id)
            elif sample_id is sample_name is None:
                sample_id = "sample_%d_id" % (i,)
                sample_name = "sample_%d" % (i,)
            sample['id'] = sample_id
            sample['name'] = sample_name

        sample_entries = [
            self.Sample(**sample) for sample in samples
        ]

        self.SampleList(sample_entries).write(self)

    def run(self, id=None, instrument_configuration=None, source_file=None, start_time=None, sample=None):
        """Begins the `<run>` section of the document, describing a single
        sample run.

        Parameters
        ----------
        id : str, optional
            The unique identifier for this element
        instrument_configuration : str, optional
            The id string for the default `InstrumentConfiguration` for this
            sample
        source_file : str, optional
            The id string for the source file used to produce this data
        start_time : str, optional
            A string encoding the date and time the sample was acquired
        sample: str, optional
            The id string for the sample used to produce this data

        Returns
        -------
        RunSection
        """
        kwargs = {}
        if start_time is not None:
            kwargs['startTimeStamp'] = start_time
        if instrument_configuration is None:
            keys = list(self.context['InstrumentConfiguration'].keys())
            if keys:
                instrument_configuration = keys[0]
            else:
                instrument_configuration = None
        self.default_instrument_configuration = instrument_configuration
        return RunSection(
            self.writer, self.context, id=id,
            instrument_configuration=instrument_configuration,
            source_file=source_file,
            sample=sample, **kwargs)

    def spectrum_list(self, count, data_processing_method=None):
        if data_processing_method is None:
            dp_map = self.context['DataProcessing']
            try:
                data_processing_method = list(dp_map.keys())[0]
            except IndexError:
                warnings.warn("No Data Processing method found. mzML file may not be fully standard-compliant")
        return SpectrumListSection(
            self.writer, self.context, count=count,
            data_processing_method=data_processing_method)

    def chromatogram_list(self, count, data_processing_method=None):
        return ChromatogramListSection(
            self.writer, self.context, count=count,
            data_processing_method=data_processing_method)

    def write_spectrum(self, mz_array=None, intensity_array=None, charge_array=None, id=None,
                       polarity='positive scan', centroided=True, precursor_information=None,
                       scan_start_time=None, params=None, compression=COMPRESSION_ZLIB,
                       encoding=None, other_arrays=None, scan_params=None, scan_window_list=None,
                       instrument_configuration_id=None):
        if encoding is None:
            {MZ_ARRAY: np.float64}
        if params is None:
            params = []
        else:
            params = list(params)
        if scan_params is None:
            scan_params = []
        else:
            scan_params = list(scan_params)
        if other_arrays is None:
            other_arrays = []
        if scan_window_list is None:
            scan_window_list = []
        else:
            scan_window_list = list(scan_window_list)

        if isinstance(encoding, Mapping):
            encoding = defaultdict(lambda: np.float32, encoding)
        else:
            # create new variable to capture in closure
            _encoding = encoding
            encoding = defaultdict(lambda: _encoding)

        if polarity is not None:
            if isinstance(polarity, int):
                if polarity > 0:
                    polarity = 'positive scan'
                elif polarity < 0:
                    polarity = 'negative scan'
                else:
                    polarity = None
            elif 'positive' in polarity:
                polarity = 'positive scan'
            elif 'negative' in polarity:
                polarity = 'negative scan'
            else:
                polarity = None

            if polarity not in params and polarity is not None:
                params.append(polarity)

        if centroided:
            peak_mode = "centroid spectrum"
        else:
            peak_mode = 'profile spectrum'
        params.append(peak_mode)

        array_list = []
        default_array_length = len(mz_array)
        if mz_array is not None:
            mz_array_tag = self._prepare_array(
                mz_array, encoding=encoding[MZ_ARRAY], compression=compression, array_type=MZ_ARRAY)
            array_list.append(mz_array_tag)

        if intensity_array is not None:
            intensity_array_tag = self._prepare_array(
                intensity_array, encoding=encoding[INTENSITY_ARRAY], compression=compression,
                array_type=INTENSITY_ARRAY)
            array_list.append(intensity_array_tag)

        if charge_array is not None:
            charge_array_tag = self._prepare_array(
                charge_array, encoding=encoding[CHARGE_ARRAY], compression=compression,
                array_type=CHARGE_ARRAY)
            array_list.append(charge_array_tag)
        for array_type, array in other_arrays:
            if array_type is None:
                raise ValueError("array type can't be None")
            array_tag = self._prepare_array(
                array, encoding=encoding[array_type], compression=compression, array_type=array_type,
                default_array_length=default_array_length)
            array_list.append(array_tag)
        array_list_tag = self.BinaryDataArrayList(array_list)

        if precursor_information is not None:
            precursor_list = self._prepare_precursor_information(
                **precursor_information)
        else:
            precursor_list = None

        if scan_start_time is not None:
            if isinstance(scan_start_time, numbers.Number):
                scan_params.append({"name": "scan start time",
                                    "value": scan_start_time,
                                    "unitName": 'minute'})
            else:
                scan_params.append(scan_start_time)
        if self.default_instrument_configuration == instrument_configuration_id:
            instrument_configuration_id = None
        scan = self.Scan(scan_window_list=scan_window_list, params=scan_params,
                         instrument_configuration_ref=instrument_configuration_id)
        scan_list = self.ScanList([scan], params=["no combination"])

        index = self.spectrum_count
        self.spectrum_count += 1
        spectrum = self.Spectrum(
            index, array_list_tag, scan_list=scan_list, params=params, id=id,
            default_array_length=default_array_length,
            precursor_list=precursor_list)
        spectrum.write(self.writer)

    def write_chromatogram(self, time_array, intensity_array, id=None,
                           chromatogram_type="selected ion current",
                           precursor_information=None, params=None,
                           compression=COMPRESSION_ZLIB, encoding=32, other_arrays=None):
        if params is None:
            params = []
        else:
            params = list(params)

        if isinstance(encoding, Mapping):
            encoding = defaultdict(lambda: np.float32, encoding)
        else:
            # create new variable to capture in closure
            _encoding = encoding
            encoding = defaultdict(lambda: _encoding)

        if other_arrays is None:
            other_arrays = []
        array_list = []

        if precursor_information is not None:
            precursor = self._prepare_precursor_information(**precursor_information)[0]
        else:
            precursor = None

        default_array_length = len(time_array)
        if time_array is not None:
            time_array_tag = self._prepare_array(
                time_array, encoding=encoding[TIME_ARRAY], compression=compression,
                array_type=TIME_ARRAY)
            array_list.append(time_array_tag)

        if intensity_array is not None:
            intensity_array_tag = self._prepare_array(
                intensity_array, encoding=encoding[INTENSITY_ARRAY], compression=compression,
                array_type=INTENSITY_ARRAY)
            array_list.append(intensity_array_tag)

        for array_type, array in other_arrays:
            array_tag = self._prepare_array(
                array, encoding=encoding[array_type], compression=compression, array_type=array_type,
                default_array_length=default_array_length)
            array_list.append(array_tag)
        params.append(chromatogram_type)
        array_list_tag = self.BinaryDataArrayList(array_list)
        index = self.chromatogram_count
        self.chromatogram_count += 1
        chromatogram = self.Chromatogram(
            index=index, binary_data_list=array_list_tag,
            precursor=precursor,
            default_array_length=default_array_length,
            id=id, params=params)
        chromatogram.write(self.writer)

    def _prepare_array(self, array, encoding=32, compression=COMPRESSION_ZLIB,
                       array_type=None, default_array_length=None):
        if isinstance(encoding, numbers.Number):
            _encoding = int(encoding)
        else:
            _encoding = encoding
        dtype = encoding_map[_encoding]
        array = np.array(array, dtype=dtype)
        encoded_binary = encode_array(
            array, compression=compression, dtype=dtype)
        binary = self.Binary(encoded_binary)
        if default_array_length is not None and len(array) != default_array_length:
            override_length = True
        else:
            override_length = False
        params = []
        if array_type is not None:
            params.append(array_type)
            if array_type not in ARRAY_TYPES:
                params.append(NON_STANDARD_ARRAY)
        params.append(compression_map[compression])
        params.append(dtype_to_encoding[dtype])
        encoded_length = len(encoded_binary)
        return self.BinaryDataArray(
            binary, encoded_length,
            array_length=(len(array) if override_length else None),
            params=params)

    def _prepare_precursor_information(self, mz, intensity, charge, scan_id, activation=None,
                                       isolation_window_args=None, params=None):
        if params is None:
            params = []
        if activation is not None:
            activation = self.Activation(activation)
        ion = self.SelectedIon(mz, intensity, charge, params=params)
        ion_list = self.SelectedIonList([ion])
        if isolation_window_args is not None:
            isolation_window_tag = self.IsolationWindow(**isolation_window_args)
        else:
            isolation_window_tag = None
        precursor = self.Precursor(
            ion_list,
            activation=activation,
            isolation_window=isolation_window_tag,
            spectrum_reference=scan_id)
        precursor_list = self.PrecursorList([precursor])
        return precursor_list

    def format(self, outfile=None, index=True, indexer=None):
        if indexer is None:
            indexer = MzMLIndexer
        if not index:
            return super(MzMLWriter, self).format(outfile=outfile)
        # if self.outfile.closed:
        #     fh = self.outfile.name
        # else:
        #     fh = self.outfile
        try:
            if self.outfile.isatty():
                return
        except (AttributeError, ValueError):
            pass
        if hasattr(self.outfile, 'name'):
            fh = self.outfile.name
        else:
            fh = self.outfile
        opener = compression.get(fh)
        indexer = indexer(fh)
        try:
            indexer.build()
            indexer.overwrite(opener)
        except MemoryError:
            pass



import pkg_resources

from lxml import etree


def get_xsd(name):
    return pkg_resources.resource_stream(__name__, "xsd/%s" % name)


schemas = {
    'mzML': ('mzML1.1.0.xsd'),
    'indexedmzML': ('mzML1.1.2_idx.xsd'),
    'MzIdentML': ('mzIdentML1.2.0.xsd'),
}


def get_schema(name):
    schema_name = schemas[name]
    tree = etree.parse(get_xsd(schema_name))
    return etree.XMLSchema(tree)


def validate(path):
    tree = etree.parse(path)
    root = tree.getroot()
    parts = root.tag.split("}", 1)
    if len(parts) == 1:
        name = parts[0]
    else:
        name = parts[1]
    try:
        schema = get_schema(name)
    except KeyError:
        raise KeyError("Could not locate a schema for %r" % (name,))
    result = schema.validate(tree)
    return result, schema



import warnings
from datetime import datetime
try:
    from collections import Mapping, Iterable
except ImportError:
    from collections.abc import Mapping, Iterable
from numbers import Number

import numpy as np

from ..xml import _element, element, TagBase, CV
from ..document import (
    ComponentBase as _ComponentBase, NullMap, ComponentDispatcherBase,
    ParameterContainer, IDParameterContainer)
from .binary_encoding import (dtype_to_encoding, compression_map, encode_array)
from .utils import ensure_iterable


class MzML(TagBase):
    type_attrs = {
        "xmlns": "http://psi.hupo.org/ms/mzml",
        "version": "1.1.0",
        "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "xsi:schemaLocation": "http://psi.hupo.org/ms/mzml http://psidev.info/files/ms/mzML/xsd/mzML1.1.0.xsd"
    }

    def __init__(self, **attrs):
        super(MzML, self).__init__("mzML", **attrs)


_COMPONENT_NAMESPACE = 'mzml'
_xmlns = "http://psidev.info/psi/pi/mzML/1.1"


class ComponentDispatcher(ComponentDispatcherBase):
    def __init__(self, *args, **kwargs):
        super(ComponentDispatcher, self).__init__(
            *args, component_namespace=_COMPONENT_NAMESPACE, **kwargs)


class ComponentBase(_ComponentBase):
    component_namespace = _COMPONENT_NAMESPACE


default_cv_list = [
    CV(id="PSI-MS",
       uri=("https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo"),
       full_name="PSI-MS"),
    CV(id="UO",
       uri="http://ontologies.berkeleybop.org/uo.obo",
       full_name="UNIT-ONTOLOGY"),
]


class GenericCollection(ComponentBase):
    def __init__(self, tag_name, members, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, count=len(self.members))

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            for member in self.members:
                member.write(xml_file)

    def __len__(self):
        return len(self.members)

    def __iter__(self):
        return iter(self.members)

    def __getitem__(self, i):
        return self.members[i]


class IDGenericCollection(GenericCollection):
    def __init__(self, tag_name, members, id, context=NullMap):
        self.members = members
        self.tag_name = tag_name
        self.element = _element(tag_name, id=id, count=len(self.members))
        context[tag_name][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            for member in self.members:
                member.write(xml_file)


# --------------------------------------------------
# File Metadata


class FileContent(ComponentBase):
    def __init__(self, params=None, context=NullMap, **kwargs):
        self.params = self.prepare_params(params, **kwargs)
        self.element = _element("fileContent")
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file):
            self.write_params(xml_file)


class SourceFileList(GenericCollection):
    def __init__(self, members, context):
        super(SourceFileList, self).__init__("sourceFileList", members, context)


class SourceFile(ComponentBase):
    def __init__(self, location=None, name=None, id=None, params=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        params.extend(kwargs.items())
        self.location = location
        self.name = name
        self.element = _element("sourceFile", location=location, id=id, name=name)
        self.params = params
        self.context = context
        context["SourceFile"][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            self.write_params(xml_file)


class FileDescription(ComponentBase):
    def __init__(self, content, source_files, contacts=None, context=NullMap):
        if not isinstance(content, FileContent):
            content = FileContent(content, context=context)
        if not isinstance(source_files, SourceFileList):
            if len(source_files) > 0 and not isinstance(source_files[0], SourceFile):
                source_files = [SourceFile(context=context, **f) for f in source_files]
            source_files = SourceFileList(source_files, context=context)
        contacts = ensure_iterable(contacts)
        if contacts and not isinstance(contacts[0], Contact):
            contacts = [Contact.ensure(contact, context=context) for contact in contacts]
        self.content = content
        self.source_files = source_files
        self.contacts = contacts
        self.context = context
        self.element = _element("fileDescription")

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            self.content.write(xml_file)
            self.source_files.write(xml_file)
            for contact in self.contacts:
                contact.write(xml_file)


# --------------------------------------------------
# ParamGroups

class ReferenceableParamGroupList(GenericCollection):
    def __init__(self, members, context):
        super(ReferenceableParamGroupList, self).__init__(
            "referenceableParamGroupList", members, context)


class ReferenceableParamGroup(IDParameterContainer):
    def __init__(self, params=None, id=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        super(ReferenceableParamGroup, self).__init__(
            "referenceableParamGroup", params, dict(id=id), context=context, **kwargs)
        self.id = self.element.id
        self.context = context
        self.context["ReferenceableParamGroup"][id] = self.element.id

# --------------------------------------------------
# Sample Metadata


class SampleList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(SampleList, self).__init__("sampleList", members, context)


class Sample(IDParameterContainer):
    def __init__(self, name, params=None, id=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        params.extend(kwargs.items())
        super(Sample, self).__init__("sample", params, dict(name=name, id=id), context)
        self.name = name
        self.id = self.element.id
        context["Sample"][id] = self.element.id

# --------------------------------------------------
# Software Processing Metadata


class SoftwareList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(SoftwareList, self).__init__("softwareList", members, context)


class Software(IDParameterContainer):
    def __init__(self, id=None, version="0.0", params=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        params.extend(kwargs.items())
        super(Software, self).__init__(
            "software", params, dict(id=id, version=version), context=context)
        self.version = version
        self.id = self.element.id
        context['Software'][id] = self.element.id

    @property
    def with_id(self):
        return True


# --------------------------------------------------
# Scan Settings Metadata


class ScanSettingsList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(ScanSettingsList, self).__init__("scanSettingsList", members, context)


class ScanSettings(ComponentBase):
    def __init__(self, id=None, source_file_references=None, target_list=None, params=None, context=NullMap):
        if source_file_references is None:
            source_file_references = []
        if target_list is None:
            target_list = []
        if params is None:
            params = []
        self.params = params
        self.source_file_references = source_file_references
        self.target_list = target_list
        self.element = _element("scanSettings", id=id)
        self.id = self.element.id
        self.context = context
        context['ScanSettings'][id] = self.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            source_refs = GenericCollection(
                "sourceFileRefList",
                [_element("sourceFileRef", ref=i) for i in self.source_file_references])
            if len(source_refs):
                source_refs.write(xml_file)
            # TODO handle targetList and targets
            self.write_params(xml_file)


# --------------------------------------------------
# Instrument Configuration Metadata

class InstrumentConfigurationList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(InstrumentConfigurationList, self).__init__("instrumentConfigurationList", members, context)


class InstrumentConfiguration(ComponentBase):
    def __init__(self, id, component_list=None, params=None,
                 software_reference=None, context=NullMap):
        if params is None:
            params = []
        if not isinstance(component_list, ComponentList):
            component_list = ComponentList(component_list, context=context)
        self.params = params
        self.software_reference = software_reference
        self._software_reference = context['Software'][software_reference]
        self.component_list = component_list
        self.element = _element(
            "instrumentConfiguration", id=id)
        self.context = context
        self.context['InstrumentConfiguration'][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            self.write_params(xml_file)
            if self.component_list is not None:
                self.component_list.write(xml_file)
            if self.software_reference is not None:
                _element("softwareRef", ref=self._software_reference).write(xml_file)


class ComponentList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(ComponentList, self).__init__("componentList", members, context)

    @classmethod
    def build(cls, members, context=NullMap, type_key='type'):
        components = []
        for component in sorted(members, key=lambda x: int(x['order'])):
            if component[type_key] == 'source':
                components.append(
                    Source(int(component['order']), component['params'], context))
            elif component[type_key] == 'analyzer':
                components.append(
                    Analyzer(int(component['order']), component['params'], context))
            elif component[type_key] == 'detector':
                components.append(
                    Detector(int(component['order']), component['params'], context))
            else:
                raise KeyError("Unknown component %s" % component[type_key])
        return cls(components, context=context)


class Source(ParameterContainer):
    def __init__(self, order, params=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        params.extend(kwargs.items())
        super(Source, self).__init__(
            "source", params, dict(order=order), context=context)
        self.order = order


class Analyzer(ParameterContainer):
    def __init__(self, order, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(Analyzer, self).__init__("analyzer", params, dict(order=order), context=context)
        self.order = order


class Detector(ParameterContainer):
    def __init__(self, order, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        super(Detector, self).__init__("detector", params, dict(order=order), context=context)
        self.order = order


# --------------------------------------------------
# Data Processing Metadata


class DataProcessingList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(DataProcessingList, self).__init__("dataProcessingList", members, context)


class DataProcessing(ComponentBase):
    def __init__(self, processing_methods=None, id=None, context=NullMap):
        if processing_methods is None:
            processing_methods = []
        if processing_methods and not isinstance(processing_methods[0], ProcessingMethod):
            processing_methods = [ProcessingMethod(context=context, **m) for m in (
                processing_methods)]
        self.processing_methods = processing_methods
        self.element = _element("dataProcessing", id=id)
        self.context = context
        context['DataProcessing'][id] = self.element.id

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            for method in self.processing_methods:
                method.write(xml_file)


class ProcessingMethod(ParameterContainer):
    def __init__(self, order, software_reference, params=None, context=NullMap, **kwargs):
        params = self.prepare_params(params, **kwargs)
        self.order = order
        self.software_reference = software_reference
        self._software_reference = context['Software'][software_reference]
        self.element = _element("processingMethod", order=order, softwareRef=self._software_reference)
        self.params = params
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            self.write_params(xml_file)

# --------------------------------------------------
# Spectral and Chromatographic Data Storage


class SpectrumList(ComponentBase):
    def __init__(self, members, default_data_processing_reference, context=NullMap):
        self.members = members
        self.default_data_processing_reference = default_data_processing_reference
        self._default_data_processing_reference = context["DataProcessing"][default_data_processing_reference]
        self.element = _element(
            "spectrumList", count=len(self.members), defaultDataProcessingRef=self._default_data_processing_reference)
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            for member in self.members:
                member.write(xml_file)


class Spectrum(ComponentBase):
    def __init__(self, index, binary_data_list=None, scan_list=None, precursor_list=None, product_list=None,
                 default_array_length=None, source_file_reference=None, data_processing_reference=None,
                 id=None, params=None, context=NullMap, **kwargs):
        if params is None:
            params = []
        self.index = index
        self.binary_data_list = binary_data_list
        self.scan_list = scan_list
        self.data_processing_reference = data_processing_reference
        self._data_processing_reference = context["DataProcessing"][data_processing_reference]
        self.precursor_list = precursor_list
        self.product_list = product_list
        self.default_array_length = default_array_length
        self.source_file_reference = source_file_reference
        self._source_file_reference = context["SourceFile"][source_file_reference]
        self.element = _element(
            "spectrum", id=id, index=index, sourceFileRef=self._source_file_reference,
            defaultArrayLength=self.default_array_length, dataProcessingRef=self._data_processing_reference)
        self.context = context
        self.context["Spectrum"][id] = self.element.id
        self.params = self.prepare_params(params, **kwargs)
        self._check_params()

    def _check_params(self):
        ms_level = None
        spectrum_type = None
        for param in self.params:
            param = self.context.param(param)
            try:
                term = self.context.term(param.accession)
            except (AttributeError, KeyError):
                continue
            if term.id == 'MS:1000511':
                ms_level = int(param.value)
            elif term.id == 'MS:1000579' or term.id == 'MS:1000580':
                spectrum_type = term.name
        if ms_level is not None and spectrum_type is None:
            spectrum_type = "MS:1000579" if ms_level == 1 else "MS:1000580"
            self.params.append(self.context.param(spectrum_type))
        elif spectrum_type is not None and ms_level is None:
            if spectrum_type == 'MS:1000579':
                self.params.append(self.context.param(accession='MS:1000511', value=1))
            else:
                raise ValueError("A spectrum without MS:100511 'ms level' and cannot "
                                 "be determined from other parameters")

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            self.write_params(xml_file)
            if self.scan_list is not None:
                self.scan_list.write(xml_file)
            if self.precursor_list is not None:
                self.precursor_list.write(xml_file)
            if self.product_list is not None:
                self.product_list.write(xml_file)
            self.binary_data_list.write(xml_file)


class Run(ComponentBase):
    def __init__(self, default_instrument_configuration_reference, spectrum_list=None, chromatogram_list=None, id=None,
                 default_source_file_reference=None, sample_reference=None, start_time_stamp=None, params=None,
                 context=NullMap):
        if params is None:
            params = []
        if spectrum_list is None:
            spectrum_list = []
        if chromatogram_list is None:
            chromatogram_list = []
        self.params = params
        self.spectrum_list = spectrum_list
        self.chromatogram_list = chromatogram_list
        self.sample_reference = sample_reference
        self._sample_reference = context["Sample"][sample_reference]
        self.default_instrument_configuration_reference = default_instrument_configuration_reference
        self._default_instrument_configuration_reference = context["InstrumentConfiguration"][
            default_instrument_configuration_reference]
        self.default_source_file_reference = default_source_file_reference
        self._default_source_file_reference = context["SourceFile"][default_source_file_reference]
        self.start_time_stamp = start_time_stamp
        self.element = _element(
            "run", id=id, defaultInstrumentConfigurationRef=self._default_instrument_configuration_reference,
            defaultSourceFileRef=self._default_source_file_reference, sampleRef=self._sample_reference,
            startTimeStamp=start_time_stamp)
        self.context = context


class BinaryDataArrayList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(BinaryDataArrayList, self).__init__("binaryDataArrayList", members, context)


class BinaryDataArray(ComponentBase):
    def __init__(self, binary, encoded_length, data_processing_reference=None, array_length=None,
                 params=None, context=NullMap):
        if params is None:
            params = []
        self.encoded_length = encoded_length
        self.data_processing_reference = data_processing_reference
        self._data_processing_reference = context["DataProcessing"][data_processing_reference]
        self.array_length = array_length
        self.params = params
        self.binary = binary

        self.element = _element(
            "binaryDataArray", arrayLength=array_length, encodedLength=encoded_length,
            dataProcessingRef=self._data_processing_reference)
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            self.write_params(xml_file)
            self.binary.write(xml_file)

    @classmethod
    def from_array(cls, data_array, compression=None, params=None, context=None):
        if params is None:
            params = []
        if context is None:
            context = NullMap
        compression = compression_map[compression]
        encoded_binary = encode_array(
            data_array, compression=compression, dtype=data_array.dtype.type)
        binary = Binary(encoded_binary)
        array_length = len(data_array)
        params.append(compression_map[compression])
        params.append(dtype_to_encoding[data_array.dtype.type])
        encoded_length = len(encoded_binary)
        inst = cls(
            binary, encoded_length,
            array_length=array_length,
            params=params, context=context)
        return inst


class Binary(ComponentBase):
    def __init__(self, encoded_array, context=NullMap):
        self.encoded_array = encoded_array
        self.context = context
        self.element = _element("binary", text=encoded_array)

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            xml_file.write(self.encoded_array)


class ScanList(ComponentBase):
    def __init__(self, members, params=None, context=NullMap):
        if params is None:
            params = []
        self.members = members
        self.params = params
        self.element = _element("scanList", count=len(self.members))
        self.context = context

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            self.write_params(xml_file)
            for member in self.members:
                member.write(xml_file)


class Scan(ComponentBase):
    def __init__(self, scan_window_list=None, instrument_configuration_ref=None, params=None,
                 context=NullMap, **kwargs):
        if scan_window_list is None:
            scan_window_list = ScanWindowList([], context)
        elif not isinstance(scan_window_list, ScanWindowList):
            scan_window_list = ScanWindowList(scan_window_list, context)
        self.instrument_configuration_ref = instrument_configuration_ref
        self._instrument_configuration_ref = context['InstrumentConfiguration'].get(instrument_configuration_ref)
        self.params = self.prepare_params(params, **kwargs)
        self.scan_window_list = scan_window_list
        self.element = _element("scan")
        if self._instrument_configuration_ref is not None:
            self.element.attrs['instrumentConfigurationRef'] = self._instrument_configuration_ref
        self.context = context

    def write(self, xml_file):
        with self.element(xml_file, with_id=False):
            self.write_params(xml_file)
            if len(self.scan_window_list) > 0:
                self.scan_window_list.write(xml_file)


class ScanWindowList(GenericCollection):
    def __init__(self, members, context=NullMap):
        components = []
        for member in members:
            if isinstance(member, (list, tuple)):
                components.append(ScanWindow(*member, context=context))
            elif isinstance(member, Mapping):
                components.append(ScanWindow(context=context, **member))
            else:
                components.append(member)
        super(ScanWindowList, self).__init__('scanWindowList', components, context=context)


class ScanWindow(ComponentBase):
    def __init__(self, lower, upper, params=None, context=NullMap):
        if params is None:
            params = []
        self.lower = lower
        self.upper = upper
        self.params = params
        self.element = _element("scanWindow")
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            self.context.param(name="scan window lower limit", value=self.lower,
                               unit_name='m/z', unit_accession="MS:1000040", unit_cv_ref='MS')(xml_file)
            self.context.param(name="scan window upper limit", value=self.upper,
                               unit_name='m/z', unit_accession="MS:1000040", unit_cv_ref='MS')(xml_file)
            self.write_params(xml_file)


class IsolationWindow(ComponentBase):
    def __init__(self, lower, target, upper, params=None, context=NullMap):
        if params is None:
            params = []
        self.target = target
        self.lower = lower
        self.upper = upper
        self.element = _element("isolationWindow")
        self.context = context
        self.params = params

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            self.context.param(name="isolation window target m/z", value=self.target,
                               unit_name='m/z', unit_accession="MS:1000040",
                               unit_cv_ref="MS")(xml_file)
            self.context.param(name="isolation window lower offset", value=self.lower,
                               unit_name='m/z', unit_accession="MS:1000040",
                               unit_cv_ref="MS")(xml_file)
            self.context.param(name="isolation window upper offset", value=self.upper,
                               unit_name='m/z', unit_accession="MS:1000040",
                               unit_cv_ref="MS")(xml_file)
            self.write_params(xml_file)


class PrecursorList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(PrecursorList, self).__init__('precursorList', members, context=context)


class Precursor(ComponentBase):
    def __init__(self, selected_ion_list, activation, isolation_window=None, spectrum_reference=None, context=NullMap):
        if isolation_window is not None:
            if isinstance(isolation_window, (tuple, list)):
                isolation_window = IsolationWindow(*isolation_window, context=context)
            elif isinstance(isolation_window, Mapping):
                isolation_window = IsolationWindow(context=context, **isolation_window)
        self.selected_ion_list = selected_ion_list
        self.activation = activation
        self.isolation_window = isolation_window
        self.spectrum_reference = spectrum_reference
        self._spectrum_reference = context["Spectrum"][spectrum_reference]
        self.element = _element("precursor", spectrumRef=self._spectrum_reference)
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            if self.isolation_window is not None:
                self.isolation_window.write(xml_file)
            self.selected_ion_list.write(xml_file)
            if self.activation is not None:
                self.activation.write(xml_file)


class ProductList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(ProductList, self).__init__('productList', members, context=context)


class Product(ComponentBase):
    def __init__(self, isolation_window=None, spectrum_reference=None, context=NullMap):
        if isolation_window is not None:
            if isinstance(isolation_window, (tuple, list)):
                isolation_window = IsolationWindow(*isolation_window, context=context)
            elif isinstance(isolation_window, Mapping):
                isolation_window = IsolationWindow(context=context, **isolation_window)
        self.isolation_window = isolation_window
        self.element = _element("product")
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=False):
            if self.isolation_window is not None:
                self.isolation_window.write(xml_file)


class Activation(ParameterContainer):
    def __init__(self, params, context=NullMap):
        super(Activation, self).__init__("activation", params, context=context)


class SelectedIonList(GenericCollection):
    def __init__(self, members, context=NullMap):
        super(SelectedIonList, self).__init__("selectedIonList", members, context=context)


class SelectedIon(ComponentBase):
    def __init__(self, selected_ion_mz, intensity=None, charge=None, params=None, context=NullMap):
        if params is None:
            params = []
        self.selected_ion_mz = selected_ion_mz
        self.intensity = intensity
        self.charge = charge
        self.params = params
        self.element = _element("selectedIon")
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file):
            if self.selected_ion_mz is not None:
                self.context.param(name="selected ion m/z", value=self.selected_ion_mz)(xml_file)
            if self.intensity is not None:
                self.context.param(name="peak intensity", value=self.intensity)(xml_file)
            if self.charge is not None:
                if isinstance(self.charge, Number):
                    self.context.param(name="charge state", value=int(self.charge))(xml_file)
                elif isinstance(self.charge, Iterable):
                    self.context.param(name="charge state", value=' '.join(map(str, self.charge)))(xml_file)
                else:
                    warnings.warn("Invalid charge state provided (%r)" % (self.charge,))
            self.write_params(xml_file)


class Chromatogram(ComponentBase):
    def __init__(self, index, binary_data_list=None, precursor=None, product=None,
                 default_array_length=None, data_processing_reference=None, id=None, params=None,
                 context=NullMap):
        if params is None:
            params = []
        self.index = index
        self.default_array_length = default_array_length
        self.binary_data_list = binary_data_list
        self.precursor = precursor
        self.product = product
        self.data_processing_reference = data_processing_reference
        self._data_processing_reference = context["DataProcessing"][data_processing_reference]
        self.element = _element(
            "chromatogram", id=id, index=index,
            defaultArrayLength=self.default_array_length,
            dataProcessingRef=self._data_processing_reference)
        self.context = context
        self.context["Chromatogram"][id] = self.element.id
        self.params = params

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            self.write_params(xml_file)
            if self.precursor is not None:
                self.precursor.write(xml_file)
            if self.product is not None:
                self.product.write(xml_file)
            self.binary_data_list.write(xml_file)


# --------------------------------------------------
# Misc. Providence Management


DEFAULT_CONTACT_ID = "PERSON_DOC_OWNER"
DEFAULT_ORGANIZATION_ID = "ORG_DOC_OWNER"


class CVList(ComponentBase):
    def __init__(self, cv_list=None, context=NullMap):
        if cv_list is None:
            cv_list = default_cv_list
        self.cv_list = cv_list

    def write(self, xml_file):
        with element(xml_file, 'cvList', count=len(self.cv_list)):
            for member in self.cv_list:
                tag = _element(
                    "cv", id=member.id, fullName=member.full_name,
                    URI=member.uri)
                if member.version is not None:
                    tag.attrs['version'] = member.version
                if member.options:
                    tag.attrs.update(member.options)
                xml_file.write(tag.element(with_id=True))

    def __iter__(self):
        return iter(self.cv_list)


class Contact(ParameterContainer):
    def __init__(self, params, context=NullMap, **kwargs):
        super(Contact, self).__init__('contact', params, context=context, **kwargs)


class Person(ComponentBase):
    def __init__(self, first_name='first_name', last_name='last_name', id=DEFAULT_CONTACT_ID,
                 affiliation=DEFAULT_ORGANIZATION_ID, context=NullMap):
        self.first_name = first_name
        self.last_name = last_name
        self.id = id
        self.affiliation = affiliation
        self.element = _element("Person", firstName=first_name, last_name=last_name, id=id)
        context["Person"][id] = self.element.id
        self.context = context

    def write(self, xml_file):
        with self.element.element(xml_file, with_id=True):
            element(xml_file, 'Affiliation', organization_ref=self.affiliation)


class Organization(ComponentBase):
    def __init__(self, name="name", id=DEFAULT_ORGANIZATION_ID, context=NullMap):
        self.name = name
        self.id = id
        self.element = _element("Organization", name=name, id=id)
        context["Organization"][id] = self.id
        self.context = context

    def write(self, xml_file):
        xml_file.write(self.element.element())


DEFAULT_PERSON = Person()
DEFAULT_ORGANIZATION = Organization()



import re

from io import BytesIO
from collections import defaultdict, OrderedDict

try:
    from collections import Sequence, Mapping
except ImportError:
    from collections.abc import Sequence, Mapping

from hashlib import sha1

from lxml import etree

from psims import compression


class Offset(object):
    __slots__ = ['offset', 'attrs']

    def __init__(self, offset, attrs):
        self.offset = offset
        self.attrs = attrs

    def __eq__(self, other):
        return int(self) == int(other)

    def __ne__(self, other):
        return int(self) != int(other)

    def __hash__(self):
        return hash(self.offset)

    def __int__(self):
        return self.offset

    def __index__(self):
        return self.offset

    def __repr__(self):
        template = "{self.__class__.__name__}({self.offset}, {self.attrs})"
        return template.format(self=self)


class TagIndexerBase(object):
    attr_pattern = re.compile(br"(\S+)=[\"']([^\"']+)[\"']")

    def __init__(self, name, pattern):
        if isinstance(pattern, str):
            pattern = pattern.encode("utf8")
        if isinstance(pattern, bytes):
            pattern = re.compile(pattern)
        self.name = name
        self.pattern = pattern
        self.index = OrderedDict()

    def __len__(self):
        return len(self.index)

    def __iter__(self):
        return iter(self.index.items())

    def scan(self, data, distance):
        is_match = self.pattern.search(data)
        if is_match:
            attrs = dict(self.attr_pattern.findall(data))
            xid = attrs[b'id']
            offset = Offset(distance + is_match.start(), attrs)
            self.index[xid] = offset
        return bool(is_match)

    def __call__(self, data, distance):
        return self.scan(data, distance)

    def write(self, writer):
        writer.write("    <index name=\"{}\">\n".format(self.name).encode('utf-8'))
        for ref_id, index_data in self.index.items():
            writer.write('      <offset idRef="{}">{:d}</offset>\n'.format(
                ref_id, int(index_data)).encode('utf-8'))
        writer.write(b"    </index>\n")


class SpectrumIndexer(TagIndexerBase):
    def __init__(self):
        super(SpectrumIndexer, self).__init__(
            'spectrum', re.compile(b"<spectrum "))


class ChromatogramIndexer(TagIndexerBase):
    def __init__(self):
        super(ChromatogramIndexer, self).__init__(
            'chromatogram', re.compile(b"<chromatogram "))


class IndexList(Sequence):
    def __init__(self, indexers=None):
        if indexers is None:
            indexers = []
        self.indexers = list(indexers)

    def __len__(self):
        return len([ix for ix in self.indexers if len(ix) > 0])

    def __getitem__(self, i):
        return self.indexers[i]

    def __iter__(self):
        return iter(self.indexers)

    def test(self, data, distance):
        for indexer in self:
            if indexer(data, distance):
                return True
        return False

    def __call__(self, data, distance):
        return self.test(data, distance)

    def write_index_list(self, writer, distance):
        offset = distance
        n = len(self)
        writer.write("  <indexList count=\"{:d}\">\n".format(n).encode("utf-8"))
        for index in self:
            if len(index) > 0:
                index.write(writer)
        writer.write(b"  </indexList>\n")
        writer.write(b"  <indexListOffset>")
        writer.write("{:d}</indexListOffset>\n".format(offset).encode("utf-8"))

    def add(self, indexer):
        self.indexers.append(indexer)


class HashingFileBuffer(object):
    def __init__(self):
        self.buffer = BytesIO()
        self.checksum = sha1()
        self.accumulator = 0

    def write(self, data):
        self.buffer.write(data)
        self.accumulator += len(data)
        self.checksum.update(data)
        return self.accumulator


class MzMLIndexer(HashingFileBuffer):
    def __init__(self, source, pretty=True):
        self.source = source
        super(MzMLIndexer, self).__init__()
        self.indices = IndexList()
        self.indices.add(SpectrumIndexer())
        self.indices.add(ChromatogramIndexer())

    def write(self, data):
        self.indices.test(data, self.accumulator)
        return super(MzMLIndexer, self).write(data)

    def write_opening(self):
        header = (b'<?xml version="1.0" encoding="utf-8"?>\n'
                  b'<indexedmzML xmlns="http://psi.hupo.org/ms/mzml" '
                  b'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
                  b' xsi:schemaLocation="http://psi.hupo.org/ms/mzml '
                  b'http://psidev.info/files/ms/mzML/xsd/mzML1.1.2_idx.xsd">\n')
        self.write(header)

    def embed_source(self):
        tree = etree.parse(self.source)
        content = etree.tostring(tree, pretty_print=True).splitlines()
        for line in content:
            indented = b''.join((b'  ', line, b'\n'))
            self.write(indented)

    def write_index(self, index, name):
        self.write("    <index name=\"{}\">\n".format(name).encode('utf-8'))
        for ref_id, index_data in index.items():
            self.write_offset(ref_id, index_data)
        self.write(b"    </index>\n")

    def write_index_list(self):
        offset = self.accumulator
        self.indices.write_index_list(self, offset)

    def write_checksum(self):
        self.write(b"  <fileChecksum>")
        self.write(self.checksum.hexdigest().encode('utf-8'))
        self.write(b"</fileChecksum>\n")

    def write_closing(self):
        self.write(b"</indexedmzML>")

    def write_offset(self, ref_id, index_data):
        self.write('      <offset idRef="{}">{:d}</offset>\n'.format(
            ref_id, index_data).encode('utf-8'))

    def build(self):
        self.write_opening()
        self.embed_source()
        self.write_index_list()
        self.write_checksum()
        self.write_closing()

    def overwrite(self, opener=None):
        if opener is None:
            opener = compression.get(self.source)
        try:
            self.source.seek(0)
            fh = self.source
        except AttributeError:
            fh = opener(self.source, 'wb')
        fh.write(self.buffer.getvalue())
